from pyHalo.PresetModels.cdm import CDM
from pyHalo.realization_extensions import RealizationExtensions


def SIDM_core_collapse(z_lens, z_source, mass_ranges_subhalos, mass_ranges_field_halos,
        probabilities_subhalos, probabilities_field_halos, sigma_sub=0.025, log10_sigma_sub=None, log10_dNdA=None,
        log_mlow=6., log_mhigh=10., concentration_model_subhalos='DIEMERJOYCE19', kwargs_concentration_model_subhalos={},
        concentration_model_fieldhalos='DIEMERJOYCE19', kwargs_concentration_model_fieldhalos={},
        truncation_model_subhalos='TRUNCATION_GALACTICUS', kwargs_truncation_model_subhalos={},
        infall_redshift_model='HYBRID_INFALL', kwargs_infall_model={},
        truncation_model_fieldhalos='TRUNCATION_RN', kwargs_truncation_model_fieldhalos={},
        shmf_log_slope=-1.9, cone_opening_angle_arcsec=6., log_m_host=13.3,  r_tidal=0.25,
        LOS_normalization=1.0, geometry_type='DOUBLE_CONE', kwargs_cosmo=None, collapsed_halo_profile='SPL_CORE',
        kwargs_collapsed_profile={'x_core_halo': 0.05, 'x_match': 'c', 'log_slope_halo': 3.0}):

    """
    Generates realizations of SIDM given the fraction of core-collapsed halos as a function of halo mass

    :param z_lens: main deflector redshift
    :param z_source: source redshift
    :param mass_ranges_subhalos: a list of mass ranges for subhalo collapse, e.g.
    mass_ranges_subhalos = [[6, 8], [8, 9], ...]
    :param mass_ranges_field_halos: a list of mass ranges for subhalo collapse, e.g.
    mass_ranges_field_halos = [[6, 8], [8, 9], ...]
    :param probabilities_subhalos: a list of functions, or a list of the fraction of collapsed subhalos in each mass bin
    :param probabilities_field_halos: a list of functions, or a list of the fraction of collapsed field halos in each
    mass bin
    :param sigma_sub: amplitude of the subhalo mass function at 10^8 solar masses in units [# of halos / kpc^2]
    :param log10_sigma_sub: amplitude of the subhalo mass function at 10^8 solar masses in units [# of halos / kpc^2];
    this setting, as well as sigma_sub, will trigger the automatic scaling with host halo mass and redshift found by Gannon et al. (2025)
    - Note: this overrides sigma_sub, if provided
    :param log10_dNdA: amplitude of the subhalo mass function at 10^8 solar masses in units [# of halos / kpc^2], overrides sigma_sub
    :param log_mlow: log base 10 of the minimum halo mass to render
    :param log_mhigh: log base 10 of the maximum halo mass to render
    :param concentration_model_subhalos: the concentration-mass relation applied to subhalos,
    see concentration_models.py for a complete list of available models
    :param kwargs_concentration_model_subhalos: keyword arguments for the subhalo MC relation
    NOTE: keyword args returned by the load_concentration_model override these keywords with duplicate arguments
    :param concentration_model_subhalos: the concentration-mass relation applied to field halos,
    see concentration_models.py for a complete list of available models
    :param kwargs_concentration_model_fieldhalos: keyword arguments for the field halo MC relation
    NOTE: keyword args returned by the load_concentration_model override these keywords with duplicate arguments
    :param truncation_model_subhalos: the truncation model applied to subhalos, see truncation_models for a complete list
    :param kwargs_truncation_model_subhalos: keyword arguments for the truncation model applied to subhalos
    :param truncation_model_fieldhalos: the truncation model applied to field halos, see truncation_models for a
    complete list
    :param kwargs_truncation_model_fieldhalos: keyword arguments for the truncation model applied to field halos
    :param infall_redshift_model: a string that specifies that infall redshift sampling distribution, see the LensCosmo
    class for details
    :param kwargs_infall_model: keyword arguments for the infall redshift model
    :param shmf_log_slope: the logarithmic slope of the subhalo mass function pivoting around 10^8 M_sun
    :param cone_opening_angle_arcsec: the opening angle of the rendering volume in arcsec
    :param log_m_host: log base 10 of the host halo mass [M_sun]
    :param r_tidal: the core radius of the host halo in units of the host halo scale radius. Subhalos are distributed
    in 3D with a cored NFW profile with this core radius
    :param LOS_normalization: rescales the amplitude of the line-of-sight mass function
    :param geometry_type: string that specifies the geometry of the rendering volume; options include
    DOUBLE_CONE, CONE, CYLINDER
    :param kwargs_cosmo: keyword arguments that specify the cosmology, see Cosmology/cosmology.py
    :param collapsed_halo_profile: string that sets the density profile of core-collapsed halos
    currently implemented models are SPL_CORE and GNFW (see example notebook)
    :param kwargs_collapsed_profile: keyword arguments for the collapsed profile (see example notebook)
    :return: a realization of dark matter structure in SIDM
    """

    two_halo_contribution = True
    delta_power_law_index = 0.0
    subhalo_spatial_distribution = 'UNIFORM'
    cdm = CDM(z_lens, z_source, sigma_sub, log_mlow, log_mhigh, log10_sigma_sub, log10_dNdA,
              concentration_model_subhalos, kwargs_concentration_model_subhalos,
              concentration_model_fieldhalos, kwargs_concentration_model_fieldhalos,
              truncation_model_subhalos, kwargs_truncation_model_subhalos,
              truncation_model_fieldhalos, kwargs_truncation_model_fieldhalos,
              infall_redshift_model, kwargs_infall_model, subhalo_spatial_distribution,
              shmf_log_slope, cone_opening_angle_arcsec, log_m_host, r_tidal,
              LOS_normalization, two_halo_contribution, delta_power_law_index,
              geometry_type, kwargs_cosmo)
    extension = RealizationExtensions(cdm)
    index_collapsed = extension.core_collapse_by_mass(mass_ranges_subhalos, mass_ranges_field_halos,
        probabilities_subhalos, probabilities_field_halos)
    sidm = extension.add_core_collapsed_halos(index_collapsed, collapsed_halo_profile, **kwargs_collapsed_profile)
    return sidm

def SIDM_parametric(z_lens, z_source, log10_mass_ranges, log10_effective_cross_section_list, log10_subhalo_time_scaling=0.0,
        sigma_sub=0.025, log10_sigma_sub=None, log10_dNdA=None, log_mlow=6., log_mhigh=10.,
        concentration_model_subhalos='DIEMERJOYCE19', kwargs_concentration_model_subhalos={},
        concentration_model_fieldhalos='DIEMERJOYCE19', kwargs_concentration_model_fieldhalos={},
        truncation_model_subhalos='TRUNCATION_GALACTICUS', kwargs_truncation_model_subhalos={},
        infall_redshift_model='HYBRID_INFALL', kwargs_infall_model={},
        truncation_model_fieldhalos='TRUNCATION_RN', kwargs_truncation_model_fieldhalos={},
        shmf_log_slope=-1.9, cone_opening_angle_arcsec=6., log_m_host=13.3,  r_tidal=0.25,
        LOS_normalization=1.0, geometry_type='DOUBLE_CONE', kwargs_cosmo=None, add_globular_clusters=False,
                    kwargs_globular_clusters=None):
    """

    Generates realizations of SIDM halos in terms of the effective scattering cross section in different mass bins using the
    parametric model by Yang et al.

    :param z_lens: main deflector redshift
    :param z_source: source redshift
    :param log10_mass_ranges: a list of lists specifying halo mass ranges, e.g. [[6, 8], [8, 10]]
    :param log10_effective_cross_section_list: a list of effective cross sections log10(sigma) for halos in each bin, e.g. [6.0, 1.0]
    :param log10_subhalo_time_scaling: a log10(number) that makes time pass quicker (>1) or lower (<1) for subhalos relative to
    field halos
    :param sigma_sub: amplitude of the subhalo mass function at 10^8 solar masses in units [# of halos / kpc^2]
    :param log10_sigma_sub: amplitude of the subhalo mass function at 10^8 solar masses in units [# of halos / kpc^2];
    this setting, as well as sigma_sub, will trigger the automatic scaling with host halo mass and redshift found by Gannon et al. (2025)
    - Note: this overrides sigma_sub, if provided
    :param log10_dNdA: amplitude of the subhalo mass function at 10^8 solar masses in units [# of halos / kpc^2], overrides sigma_sub
    :param log_mlow: log base 10 of the minimum halo mass to render
    :param log_mhigh: log base 10 of the maximum halo mass to render
    :param concentration_model_subhalos: the concentration-mass relation applied to subhalos,
    see concentration_models.py for a complete list of available models
    :param kwargs_concentration_model_subhalos: keyword arguments for the subhalo MC relation
    NOTE: keyword args returned by the load_concentration_model override these keywords with duplicate arguments
    :param concentration_model_subhalos: the concentration-mass relation applied to field halos,
    see concentration_models.py for a complete list of available models
    :param kwargs_concentration_model_fieldhalos: keyword arguments for the field halo MC relation
    NOTE: keyword args returned by the load_concentration_model override these keywords with duplicate arguments
    :param truncation_model_subhalos: the truncation model applied to subhalos, see truncation_models for a complete list
    :param kwargs_truncation_model_subhalos: keyword arguments for the truncation model applied to subhalos
    :param truncation_model_fieldhalos: the truncation model applied to field halos, see truncation_models for a
    complete list
    :param kwargs_truncation_model_fieldhalos: keyword arguments for the truncation model applied to field halos
    :param infall_redshift_model: a string that specifies that infall redshift sampling distribution, see the LensCosmo
    class for details
    :param kwargs_infall_model: keyword arguments for the infall redshift model
    :param shmf_log_slope: the logarithmic slope of the subhalo mass function pivoting around 10^8 M_sun
    :param cone_opening_angle_arcsec: the opening angle of the rendering volume in arcsec
    :param log_m_host: log base 10 of the host halo mass [M_sun]
    :param r_tidal: the core radius of the host halo in units of the host halo scale radius. Subhalos are distributed
    in 3D with a cored NFW profile with this core radius
    :param LOS_normalization: rescales the amplitude of the line-of-sight mass function
    :param geometry_type: string that specifies the geometry of the rendering volume; options include
    DOUBLE_CONE, CONE, CYLINDER
    :param kwargs_cosmo: keyword arguments that specify the cosmology, see Cosmology/cosmology.py
    :param collapsed_halo_profile: string that sets the density profile of core-collapsed halos
    currently implemented models are SPL_CORE and GNFW (see example notebook)
    :param kwargs_collapsed_profile: keyword arguments for the collapsed profile (see example notebook)
    :param add_globular_clusters: bool; include a population of globular clusters around image positions
    :param kwargs_globular_clusters: keyword arguments for the GC population; see documentation in RealizationExtensions
    :return: a realization of dark matter structure in SIDM
    """
    two_halo_contribution = True
    delta_power_law_index = 0.0
    subhalo_spatial_distribution = 'UNIFORM'
    cdm = CDM(z_lens, z_source, sigma_sub, log_mlow, log_mhigh, log10_sigma_sub, log10_dNdA,
              concentration_model_subhalos, kwargs_concentration_model_subhalos,
              concentration_model_fieldhalos, kwargs_concentration_model_fieldhalos,
              truncation_model_subhalos, kwargs_truncation_model_subhalos,
              truncation_model_fieldhalos, kwargs_truncation_model_fieldhalos,
              infall_redshift_model, kwargs_infall_model, subhalo_spatial_distribution,
              shmf_log_slope, cone_opening_angle_arcsec, log_m_host, r_tidal,
              LOS_normalization, two_halo_contribution, delta_power_law_index,
              geometry_type, kwargs_cosmo)
    extension = RealizationExtensions(cdm)
    sidm = extension.toSIDM_from_cross_section(log10_mass_ranges,
                                                   log10_effective_cross_section_list,
                                                   log10_subhalo_time_scaling)
    if add_globular_clusters:
        ext = RealizationExtensions(sidm)
        sidm = ext.add_globular_clusters(**kwargs_globular_clusters)
    return sidm

def SIDM_parametric_fixedbins(z_lens, z_source, log10_sigma_eff_mlow_8, log10_sigma_eff_8_mhigh,
        log10_subhalo_time_scaling=0.0, sigma_sub=0.025, log10_sigma_sub=None, log10_dNdA=None, log_mlow=6., log_mhigh=10.,
        concentration_model_subhalos='DIEMERJOYCE19', kwargs_concentration_model_subhalos={},
        concentration_model_fieldhalos='DIEMERJOYCE19', kwargs_concentration_model_fieldhalos={},
        truncation_model_subhalos='TRUNCATION_GALACTICUS', kwargs_truncation_model_subhalos={},
        infall_redshift_model='HYBRID_INFALL', kwargs_infall_model={},
        truncation_model_fieldhalos='TRUNCATION_RN', kwargs_truncation_model_fieldhalos={},
        shmf_log_slope=-1.9, cone_opening_angle_arcsec=6., log_m_host=13.3,  r_tidal=0.25,
        LOS_normalization=1.0, geometry_type='DOUBLE_CONE', kwargs_cosmo=None,
        add_globular_clusters=False, kwargs_globular_clusters=None):
    """
    This function uses the SIDM_parametric model with a preset list of logarithmic mass bins 10^log_mlow - 10^7.5,
    10^7.5-10^8.5, 10^8.5-10^10. The two parameters log10_sigma_eff_mlow_8 and log10_sigma_eff_8_mhigh set the cross section
     in each mass bin. For detailed documentation for the rest of the arguments, see the documentation in the SIDM_parametric function
    :param log10_sigma_eff_mlow_8: effective cross section in the mass range 10^log_mlow - 10^8
    :param log10_sigma_eff_8_mhigh: effective cross section in the mass range 10^8 - 10^log_mhigh
    :param log10_subhalo_time_scaling: a log10(number) that makes time pass quicker (>1) or lower (<1) for subhalos relative to
    field halos
    """
    if log_mlow > 8:
        raise Exception('to use this function log_mlow must be < 8')
    log10_mass_ranges = [[log_mlow, 8.0], [8.0, log_mhigh]]
    log10_effective_cross_section_list = [log10_sigma_eff_mlow_8, log10_sigma_eff_8_mhigh]
    return SIDM_parametric(z_lens, z_source, log10_mass_ranges, log10_effective_cross_section_list,
        log10_subhalo_time_scaling, sigma_sub, log10_sigma_sub, log10_dNdA, log_mlow, log_mhigh,
        concentration_model_subhalos, kwargs_concentration_model_subhalos,
        concentration_model_fieldhalos, kwargs_concentration_model_fieldhalos,
        truncation_model_subhalos, kwargs_truncation_model_subhalos,
        infall_redshift_model, kwargs_infall_model,
        truncation_model_fieldhalos, kwargs_truncation_model_fieldhalos,
        shmf_log_slope, cone_opening_angle_arcsec, log_m_host,  r_tidal,
        LOS_normalization, geometry_type, kwargs_cosmo,
        add_globular_clusters, kwargs_globular_clusters)



from pyHalo.Halos.halo_base import Halo
import numpy as np
from pyHalo.Halos.tnfw_halo_util import tnfw_mass_fraction
from lenstronomy.LensModel.Profiles.tnfw import TNFW

class TNFWFieldHalo(Halo):
    # we use the pseudo nfw methods to normalize profile
    _pseudo_nfw = False
    """
    The base class for a truncated NFW halo
    """
    def __init__(self, mass, x, y, r3d, z,
                 sub_flag, lens_cosmo_instance, args,
                 truncation_class, concentration_class, unique_tag):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        self._lens_cosmo = lens_cosmo_instance
        self._concentration_class = concentration_class
        self._truncation_class = truncation_class
        self.tnfw_lenstronomy = TNFW()
        mdef = 'TNFW'
        super(TNFWFieldHalo, self).__init__(mass, x, y, r3d, mdef, z, sub_flag,
                                           lens_cosmo_instance, args, unique_tag)

    @classmethod
    def simple_setup(cls, mass, x, y, z, tau, lens_cosmo,
                     concentration_model='DIEMERJOYCE19'):
        """
        Creates an instance of the TNFWFieldHalo class with some default assumptions for the concentration-mass relation
        and truncation
        :param mass: halo mass
        :param x: halo x-coordinate [arcsec]
        :param y: halo y-coordinate [arcsec]
        :param z: halo redshift
        :param tau: the truncation radius in units of rs
        :param lens_cosmo: an instance of lens_cosmo class
        :param concentration_model: the concentration-mass relation model
        :return: an instance of TNFWFieldHalo class
        """
        r3d = None
        sub_flag = False
        args = {}
        from pyHalo.concentration_models import preset_concentration_models
        from pyHalo.truncation_models import truncation_models
        _c_model, _ = preset_concentration_models(concentration_model)
        _t_model, _ = truncation_models('MULTIPLE_RS')
        concentration_class = _c_model(lens_cosmo.cosmo.astropy, scatter=False)
        truncation_class = _t_model(lens_cosmo, tau)
        return TNFWFieldHalo(mass, x, y, r3d, z, sub_flag, lens_cosmo, args,
                             truncation_class, concentration_class, 1.0)

    def density_profile_3d(self, r, profile_args=None, scaling=1.0):
        """
        Computes the 3-D density profile of the halo
        :param r: distance from center of halo [kpc]
        :return: the density profile in units M_sun / kpc^3
        """
        if profile_args is None:
            c, rt = self.profile_args
        else:
            c, rt = profile_args
        rhos, rs, _ = self.lens_cosmo.NFW_params_physical(self.mass, self.c, self.z_eval)
        tau = rt / rs
        x = r / rs
        rho_nfw = rhos / x / (1 + x) ** 2
        return scaling * rho_nfw * tau ** 2 / (tau ** 2 + x ** 2)

    def mass_3d(self, rmax, profile_args=None):
        """
        Calculate the enclosed mass in 3D
        :param rmax:
        :param profile_args:
        :return:
        """
        if rmax == 'r200':
            rmax = self.nfw_params[1] * self.c
        rs = self.nfw_params[1]
        rho0 = self.nfw_params[0] * self._rescale_norm
        r_trunc = self.profile_args[1]
        return self.tnfw_lenstronomy.mass_3d(rmax, rs, rho0, r_trunc/rs)

    def density_profile_3d_lenstronomy(self, r, profile_args=None):
        """
        Computes the 3-D density profile of the halo
        :param r: distance from center of halo [kpc]
        :return: the density profile in units M_sun / kpc^3
        """
        prof = self.tnfw_lenstronomy
        kwargs_lenstronomy = self.lenstronomy_params[0][0]
        kpc_per_arcsec = self._lens_cosmo.cosmo.kpc_proper_per_asec(self.z)
        sigma_crit_mpc = self._lens_cosmo.get_sigma_crit_lensing(self.z, self._lens_cosmo.z_source)
        sigma_crit_kpc = sigma_crit_mpc * 1e-6
        factor = sigma_crit_kpc / kpc_per_arcsec
        rhos = prof.alpha2rho0(kwargs_lenstronomy['alpha_Rs'], kwargs_lenstronomy['Rs'])
        return factor*prof.density(r / kpc_per_arcsec, kwargs_lenstronomy['Rs'],
                                                    rhos,
                                                    kwargs_lenstronomy['r_trunc'])

    @property
    def lenstronomy_ID(self):
            """
            See documentation in base class (Halos/halo_base.py)
            """

            return ['TNFW']

    @property
    def c(self):
        """
        Computes the halo concentration (once)
        """

        if not hasattr(self, '_c'):
            self._c = self._concentration_class.nfw_concentration(self.mass, self.z_eval)
        return self._c

    @property
    def params_physical(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """

        if not hasattr(self, '_params_physical'):
            [_, rt] = self.profile_args
            rhos, rs, r200 = self.nfw_params
            self._params_physical = {'rhos': rhos * self._rescale_norm, 'rs': rs, 'r200': r200, 'r_trunc_kpc': rt}

        return self._params_physical

    @property
    def lenstronomy_params(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_kwargs_lenstronomy'):

            [_, rt] = self.profile_args
            rhos_kpc, rs_kpc, _ = self.nfw_params
            rhos_mpc = rhos_kpc * 1e9
            rs_mpc = rs_kpc * 1e-3
            Rs_angle, theta_Rs = self._lens_cosmo.nfw_physical2angle_fromNFWparams(rhos_mpc,
                                                                                   rs_mpc,
                                                                                   self.z)
            x, y = np.round(self.x, 4), np.round(self.y, 4)
            Rs_angle = np.round(Rs_angle, 10)
            theta_Rs = np.round(theta_Rs, 10)
            r_trunc_arcsec = rt / self._lens_cosmo.cosmo.kpc_proper_per_asec(self.z)
            kwargs = [{'alpha_Rs': self._rescale_norm * theta_Rs, 'Rs': Rs_angle,
                      'center_x': x, 'center_y': y, 'r_trunc': r_trunc_arcsec}]
            self._kwargs_lenstronomy = kwargs

        return self._kwargs_lenstronomy, None

    @property
    def profile_args(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_profile_args'):
            truncation_radius_kpc = self._truncation_class.truncation_radius_halo(self)
            self._profile_args = (self.c, truncation_radius_kpc)
        return self._profile_args

    @property
    def vmax_nfw(self):
        """
        Returns the maximum circular velocity in km/sec
        :return:
        """
        if not hasattr(self, '_vmax'):
            rhos, rs, _ = self.nfw_params
            _ = self.profile_args
            self._vmax = self._lens_cosmo.nfw_vmax(self._rescale_norm * rhos, rs)
        return self._vmax

    @property
    def pseudo_nfw(self):
        return False

class TNFWSubhalo(TNFWFieldHalo):
    """
    Defines a truncated NFW halo that is a subhalo of the host dark matter halo
    """

    @property
    def bound_mass(self):
        """
        Computes the mass inside the infall virial radius (with truncation effects included)
        :return: the mass inside r = c * r_s
        """
        if hasattr(self, '_kwargs_lenstronomy'):
            tau = self._kwargs_lenstronomy[0]['r_trunc'] / self._kwargs_lenstronomy[0]['Rs']
        else:
            params_physical = self.params_physical
            tau = params_physical['r_trunc_kpc'] / params_physical['rs']
        f = tnfw_mass_fraction(tau, self.c)
        return f * self.mass * self._rescale_norm

    @property
    def bound_mass_galacticus_definition(self):
        """
        Computes the mass inside the virial radius (with truncation effects included)
        :return: the mass inside r = c * r_s
        """
        if self._truncation_class.name in ['TruncationGalacticus', 'TruncationGalacticusKeeley24']:
            pass
        else:
            raise Exception('this method can only be called when using the TruncationGalacticus class')
        if not hasattr(self, '_mbound_galacticus_definition'):
            self._mbound_galacticus_definition = self._truncation_class.calculate_mbound(self)
        return self._mbound_galacticus_definition

    @property
    def profile_args(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_profile_args'):
            if self._truncation_class.name == 'TruncationGalacticus':
                mbound = self.bound_mass_galacticus_definition
                truncation_radius_kpc = self._truncation_class.truncation_radius_from_bound_mass(self, mbound)
            else:
                truncation_radius_kpc = self._truncation_class.truncation_radius_halo(self)
            self._profile_args = (self.c, truncation_radius_kpc)
        return self._profile_args

    @property
    def pseudo_nfw(self):
        return False



from pyHalo.Halos.halo_base import Halo
from lenstronomy.LensModel.Profiles.pseudo_double_powerlaw import PseudoDoublePowerlaw
import numpy as np

class PrompCusp(Halo):
    """
    The base class for a prompt cusp
    """
    _gamma_inner = 1.5
    _gamma_outer = 10.0
    def __init__(self, mass, x, y, r3d, z,
                 sub_flag, lens_cosmo_instance, args, truncation_class, concentration_class, unique_tag):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        self._prof = PseudoDoublePowerlaw()
        self._lens_cosmo = lens_cosmo_instance
        self._truncation_class = truncation_class
        self._concentration_class = concentration_class
        mdef = 'PSEUDO_DPL'
        super(PrompCusp, self).__init__(mass, x, y, r3d, mdef, z, sub_flag,
                                              lens_cosmo_instance, args, unique_tag)

    def density_profile_3d(self, r):
        """

        :param r:
        :return:
        """
        kpc_per_arcsec = self._lens_cosmo.cosmo.kpc_proper_per_asec(self.z)
        rho0_arcsec = self.density_normalization
        sigma_crit_mpc = self._lens_cosmo.get_sigma_crit_lensing(self.z, self._lens_cosmo.z_source)
        sigma_crit_arcsec = sigma_crit_mpc * (0.001 * kpc_per_arcsec) ** 2
        rho0_kpc = rho0_arcsec * sigma_crit_arcsec / kpc_per_arcsec ** 3
        rs_kpc = self._args['cusp_R'] * 1000
        return self._prof.density(r, rs_kpc, rho0_kpc, self._gamma_inner, self._gamma_outer)

    @property
    def lenstronomy_params(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_lenstronomy_args'):

            kpc_per_arcsec = self._lens_cosmo.cosmo.kpc_proper_per_asec(self.z)
            x, y = np.round(self.x, 4), np.round(self.y, 4)
            rho0 = self.density_normalization
            R = self._args['cusp_R'] * 0.001
            R = np.round(R / kpc_per_arcsec, 10)
            alpha_Rs = self._prof.rho02alpha(rho0, R, self._gamma_inner, self._gamma_outer)
            alpha_Rs = np.round(alpha_Rs, 10)
            self._lenstronomy_args = [{'alpha_Rs': self._rescale_norm * alpha_Rs,
                                       'Rs': R,
                                       'gamma_inner': self._gamma_inner,
                                       'center_x': x,
                                       'center_y': y,
                                      'gamma_outer': self._gamma_outer}]
        return self._lenstronomy_args, None

    @property
    def profile_args(self):
        """

        :return:
        """
        return (self.density_normalization, self._args['cusp_R'], self._args['cusp_A'])

    @property
    def lenstronomy_ID(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        return ['PSEUDO_DPL']

    @property
    def density_normalization(self):
        """
        Calculates the density normalization that conserves mass within x_match
        :return:
        """
        if not hasattr(self, '_rho0_norm'):

            kpc_per_arcsec = self._lens_cosmo.cosmo.kpc_proper_per_asec(self.z)
            sigma_crit_mpc = self._lens_cosmo.get_sigma_crit_lensing(self.z, self._lens_cosmo.z_source)
            sigma_crit_arcsec = sigma_crit_mpc * (0.001 * kpc_per_arcsec) ** 2
            R_mpc = self._args['cusp_R'] # in mpc
            cuspA = self._args['cusp_A'] # M_sun * mpc^-1.5
            R_kpc = R_mpc * 1000
            mass = 8 * np.pi / 3 * cuspA * R_mpc ** 1.5
            self._rho0_norm = mass / self._prof.mass_3d(R_kpc/kpc_per_arcsec,
                                                        R_kpc/kpc_per_arcsec,
                                                     sigma_crit_arcsec,
                                                     self._gamma_inner,
                                                     self._gamma_outer)
        return self._rho0_norm



import numpy.testing as npt
from pyHalo.Halos.lens_cosmo import LensCosmo
from pyHalo.Halos.accretion import InfallDistributionGalacticus2024, InfallDistributionHybrid
import numpy as np
import pytest
from astropy.cosmology import FlatLambdaCDM
from astropy.constants import G, c
import astropy.units as un
from pyHalo.Cosmology.cosmology import Cosmology
from lenstronomy.Cosmo.lens_cosmo import LensCosmo as LensCosmoLenstronomy


class TestLensCosmo(object):

    def setup_method(self):

        kwargs_cosmo = {'Om0': 0.2}
        self.cosmo = Cosmology(cosmo_kwargs=kwargs_cosmo)
        zlens, zsource = 0.3, 1.7
        self.lens_cosmo = LensCosmo(zlens, zsource, self.cosmo)
        self.h = self.cosmo.h

    def test_const(self):

        D_ds = self.cosmo.D_A(0.3, 1.7)
        D_d = self.cosmo.D_A_z(0.3)
        D_s = self.cosmo.D_A_z(1.7)

        c_Mpc_sec = un.Quantity(c, unit=un.Mpc / un.s)
        G_Mpc3_Msun_sec2 = un.Quantity(G, unit=un.Mpc ** 3 / un.s ** 2 / un.solMass)

        const = c_Mpc_sec ** 2 / (4 * np.pi * G_Mpc3_Msun_sec2)
        sigma_crit_mpc = const.value * D_s / (D_d * D_ds)
        sigma_crit_kpc = sigma_crit_mpc * 1000 ** -2

        npt.assert_almost_equal(self.lens_cosmo.sigma_crit_lensing/sigma_crit_mpc, 1, 4)
        npt.assert_almost_equal(self.lens_cosmo.sigma_crit_lens_kpc/sigma_crit_kpc, 1, 4)

    def test_sigma_crit_mass(self):

        area = 2.
        sigma_crit_mass = self.lens_cosmo.sigma_crit_mass(0.7, area)
        sigma_crit = self.lens_cosmo.get_sigma_crit_lensing(0.7, 1.7)
        npt.assert_almost_equal(sigma_crit_mass, sigma_crit * area)

    def test_mhm_convert(self):

        mthermal = 5.3
        mhm = self.lens_cosmo.mthermal_to_halfmode(mthermal)
        mthermal_out = self.lens_cosmo.halfmode_to_thermal(mhm)
        npt.assert_almost_equal(mthermal/mthermal_out, 1, 2)

        fsl = self.lens_cosmo.mhm_to_fsl(10**8.)
        npt.assert_array_less(fsl, 100)

    def test_nfw_definitions_wrt_lenstronomy(self):

        astropy = FlatLambdaCDM(70, 0.3)
        cosmo = Cosmology(astropy)
        zlens = 0.5
        zsource = 2.0
        lc = LensCosmoLenstronomy(zlens, zsource, astropy)
        lens_cosmo = LensCosmo(zlens, zsource, cosmo)

        rho0_lenstronomy, rs_lenstronomy, r200_lenstronomy = lc.nfwParam_physical(10 ** 8, 16.0)
        rho0, rs, r200 = lens_cosmo.nfwParam_physical(10 ** 8, 16.0, zlens)
        npt.assert_almost_equal(rho0_lenstronomy/rho0, 1, 4)
        npt.assert_almost_equal(rs/rs_lenstronomy, 1.0, 4)
        npt.assert_almost_equal(r200/r200_lenstronomy, 1.0, 4)

        rho0_kpc, rs_kpc, r200_kpc = lens_cosmo.NFW_params_physical(10 ** 8, 16.0, zlens)
        npt.assert_almost_equal(rho0_kpc, rho0*1000**-3)
        npt.assert_almost_equal(rs_kpc, rs*1000)
        npt.assert_almost_equal(r200_kpc, r200 * 1000)

    def test_infall(self):

        zlens = 0.5
        zsource = 2.0

        lens_cosmo = LensCosmo(zlens, zsource, self.cosmo, infall_redshift_model='HYBRID_INFALL')
        z_infall = lens_cosmo.z_accreted_from_zlens(10 ** 8)
        npt.assert_equal(True, z_infall > zlens)

        lens_cosmo = LensCosmo(zlens, zsource, self.cosmo, infall_redshift_model='DIRECT_INFALL')
        z_infall = lens_cosmo.z_accreted_from_zlens(10 ** 8)
        npt.assert_equal(True, z_infall > zlens)

        infall_time_model = InfallDistributionGalacticus2024
        kwargs_infall_model = {}
        lens_cosmo = LensCosmo(zlens, zsource, self.cosmo, infall_redshift_model=infall_time_model,
                               kwargs_infall_model=kwargs_infall_model)
        z_infall = lens_cosmo.z_accreted_from_zlens(10 ** 8)
        npt.assert_equal(True, z_infall > zlens)

        infall_time_model = InfallDistributionHybrid
        kwargs_infall_model = {'log_m_host': 13.0}
        lens_cosmo = LensCosmo(zlens, zsource, self.cosmo, infall_redshift_model=infall_time_model,
                               kwargs_infall_model=kwargs_infall_model)
        z_infall = lens_cosmo.z_accreted_from_zlens(10 ** 8)
        npt.assert_equal(True, z_infall > zlens)

    def test_sidm_halo_age(self):

        z, z_infall, lambda_t, zform = 0.5, 5.0, 1.0, 10.0
        halo_age = self.lens_cosmo.sidm_halo_effective_age(z, z_infall, lambda_t, zform=10.0)
        halo_age_2 = self.lens_cosmo.sidm_halo_effective_age(z, z_infall, 2*lambda_t, zform=10.0)
        npt.assert_equal(halo_age_2 > halo_age, True)

    def test_nfw_params(self):

        m, c, z = 10**8, 11.0, 0.5
        rhos, rs, r200 = self.lens_cosmo.nfwParam_physical(m,c,z)
        Rs_Angle, Theta_Rs = self.lens_cosmo.nfw_physical2angle(m,c,z)
        rs_angle, theta_rs = self.lens_cosmo.nfw_physical2angle_fromNFWparams(rhos, rs, z)
        npt.assert_almost_equal(Rs_Angle, rs_angle)
        npt.assert_almost_equal(Theta_Rs, theta_rs)

        rhos_kpc, rs_kpc, r200_kpc = self.lens_cosmo.NFW_params_physical(m, c, z)
        npt.assert_almost_equal(rhos_kpc, rhos * 1e-9)
        npt.assert_almost_equal(rs_kpc, rs * 1e3)
        rs_angle, theta_rs = self.lens_cosmo.nfw_physical2angle_fromNFWparams(rhos_kpc * 1e9,
                                                                              rs_kpc * 1e-3,
                                                                              z)
        npt.assert_almost_equal(Rs_Angle, rs_angle)
        npt.assert_almost_equal(Theta_Rs, theta_rs)

    def test_sidm_timescale(self):

        # test scaling with halo mass
        m1 = 10 ** 10
        c1 = 5
        rhos, rs, r200 = self.lens_cosmo.NFW_params_physical(m1, c1, 0.5)
        tc1 = self.lens_cosmo.sidm_collapse_timescale(rhos, rs, 10.0)
        m2 = 10 ** 8
        c2 = 5
        rhos, rs, r200 = self.lens_cosmo.NFW_params_physical(m2, c2, 0.5)
        tc2 = self.lens_cosmo.sidm_collapse_timescale(rhos, rs, 10.0)
        npt.assert_almost_equal((m2/m1) ** (-1/3), tc2/tc1, 6)

        # test scaling with concentration
        m1 = 10 ** 8
        c1 = 10
        rhos, rs, r200 = self.lens_cosmo.NFW_params_physical(m1, c1, 0.5)
        tc1 = self.lens_cosmo.sidm_collapse_timescale(rhos, rs, 10.0)
        m2 = 10 ** 8
        c2 = 20
        rhos, rs, r200 = self.lens_cosmo.NFW_params_physical(m2, c2, 0.5)
        tc2 = self.lens_cosmo.sidm_collapse_timescale(rhos, rs, 10.0)
        npt.assert_almost_equal((c2 / c1) ** (-7 / 2), tc2 / tc1, 1)

        rhos = 2.74e8
        rs = 0.141
        tc_yangetal_2023 = self.lens_cosmo.sidm_collapse_timescale(rhos, rs, 7.1)
        npt.assert_almost_equal(tc_yangetal_2023, 28.07, 2)


if __name__ == '__main__':
    pytest.main()




from pyHalo.Halos.halo_base import Halo
from lenstronomy.LensModel.Profiles.nfw import NFW as NFWLenstronomy
import numpy as np

class NFWFieldHalo(Halo):
    # we use the pseudo nfw methods to normalize profile
    _pseudo_nfw = False
    """
    The main class for an NFW field halo profile without truncation

    See the base class in Halos/halo_base.py for the required routines for any instance of a Halo class
    """

    def __init__(self, mass, x, y, r3d, z,
                 sub_flag, lens_cosmo_instance, args, truncation_class, concentration_class, unique_tag):

        """
        See documentation in base class (Halos/halo_base.py)
        """
        self._lens_cosmo = lens_cosmo_instance
        self._truncation_class = truncation_class
        self._concentration_class = concentration_class
        self._nfw_lenstronomy = NFWLenstronomy()
        mdef = 'NFW'
        super(NFWFieldHalo, self).__init__(mass, x, y, r3d, mdef, z, sub_flag,
                                            lens_cosmo_instance, args, unique_tag)

    def density_profile_3d(self, r, profile_args=None, scaling=1.0):
        """
        Computes the 3-D density profile of the halo
        :param r: distance from center of halo [kpc]
        :return: the density profile in units M_sun / kpc^3
        """
        rhos, rs, _ = self.lens_cosmo.NFW_params_physical(self.mass, self.c, self.z_eval)
        x = r/rs
        return scaling * rhos / x / (1 + x) ** 2

    @property
    def lenstronomy_ID(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """

        return ['NFW']

    @property
    def lenstronomy_params(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """

        rhos_kpc, rs_kpc, _ = self.nfw_params
        rhos_mpc = rhos_kpc * 1e9
        rs_mpc = rs_kpc * 1e-3
        Rs_angle, theta_Rs = self._lens_cosmo.nfw_physical2angle_fromNFWparams(rhos_mpc,
                                                                               rs_mpc,
                                                                               self.z)
        x, y = np.round(self.x, 4), np.round(self.y, 4)
        Rs_angle = np.round(Rs_angle, 10)
        theta_Rs = np.round(theta_Rs, 10)
        kwargs = [{'alpha_Rs': self._rescale_norm * theta_Rs, 'Rs': Rs_angle,
                  'center_x': x, 'center_y': y}]
        return kwargs, None

    @property
    def c(self):
        """
        Computes the halo concentration (once)
        """

        if not hasattr(self, '_c'):
            self._c = self._concentration_class.nfw_concentration(self.mass, self.z_eval)
        return self._c

    @property
    def profile_args(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_profile_args'):
            concentration = self.c
            self._profile_args = (concentration)
        return self._profile_args

    @property
    def vmax_nfw(self):
        """
        Returns the maximum circular velocity in km/sec
        :return:
        """
        if not hasattr(self, '_vmax'):
            rhos, rs, _ = self.nfw_params
            _ = self.profile_args
            self._vmax = self._lens_cosmo.nfw_vmax(self._rescale_norm * rhos, rs)
        return self._vmax

    def mass_3d(self, rmax, profile_args=None):
        """
        Calculate the enclosed mass in 3D
        :param rmax:
        :param profile_args:
        :return:
        """
        if rmax == 'r200':
            rmax = self.nfw_params[1] * self.c
        rs = self.nfw_params[1]
        rho0 = self.nfw_params[0] * self._rescale_norm
        return self._nfw_lenstronomy.mass_3d(rmax, rs, rho0)

class NFWSubhhalo(NFWFieldHalo):

    """
    The main class for an NFW subhalo profile without truncation

    See the base class in Halos/halo_base.py for the required routines for any instance of a Halo class
    """



import numpy.testing as npt
from lenstronomy.Cosmo.lens_cosmo import LensCosmo as LensCosmoLenstronomy
from pyHalo.Halos.HaloModels.TNFW import TNFWSubhalo, TNFWFieldHalo
from pyHalo.Halos.concentration import ConcentrationDiemerJoyce
from pyHalo.Halos.tidal_truncation import TruncationRoche
from astropy.cosmology import FlatLambdaCDM
from pyHalo.Halos.lens_cosmo import LensCosmo
from pyHalo.Cosmology.cosmology import Cosmology
import pytest
import numpy as np
from pyHalo.PresetModels.cdm import CDM
from colossus.cosmology import cosmology
from colossus.halo import mass_so

class TestTNFWHalos(object):

    def setup_method(self):

        astropy = FlatLambdaCDM(70.0, 0.3)
        cosmo = Cosmology(astropy_instance=astropy)
        self.zhalo = 0.5
        self.zsource = 2.0
        self.lens_cosmo = LensCosmo(self.zhalo, self.zsource, cosmo)
        self.truncation_class = TruncationRoche(None, 100000000.0)
        self.concentration_class = ConcentrationDiemerJoyce(self.lens_cosmo, scatter=False)
        self.lclenstronomy = LensCosmoLenstronomy(self.zhalo, self.zsource, astropy)

    def test_rhos_rs_eval(self):
        """
        Test that rhos and rs are evaluated correctly for subhalos (meaning at infall)
        credit Charles Gannon
        """

        z_lens = 0.5
        z_source = 2
        cone_opening_angle_arcsec = 4
        sigma_sub = 1
        log_mlow, log_mhigh = 8, 13
        realization_pyhalo = CDM(
            z_lens,
            z_source,
            cone_opening_angle_arcsec=cone_opening_angle_arcsec,
            LOS_normalization=0,
            log_mlow=log_mlow,
            log_mhigh=log_mhigh,
            log_m_host=13,
            truncation_model_subhalos='TRUNCATION_GALACTICUS',
            log10_sigma_sub=sigma_sub,
            mass_threshold_sis=10**13
        )
        # Colossus
        cosmo = cosmology.setCosmology("planck18")
        rs_pyh = np.asarray([h.params_physical["rs"] for h in realization_pyhalo.halos])
        m_pyh = np.asarray([h.mass for h in realization_pyhalo.halos])
        z_pyh = np.asarray([h._z_infall for h in realization_pyhalo.halos])
        c_pyh = np.asarray([h.c for h in realization_pyhalo.halos])
        rv_pyh_col = [mass_so.M_to_R(m * cosmo.h, z, "200c") / cosmo.h for m, z in zip(m_pyh, z_pyh)]
        c_colossus = rv_pyh_col / rs_pyh
        npt.assert_almost_equal(c_pyh/c_colossus, np.ones_like(c_pyh),2)

    def test_simple_setup(self):
        mass = 10 ** 8
        x = 0.0
        y = 0.0
        z = 0.5
        tau = 2.2
        halo = TNFWFieldHalo.simple_setup(mass, x, y, z, tau, self.lens_cosmo)
        kwargs, _ = halo.lenstronomy_params
        npt.assert_almost_equal(kwargs[0]['r_trunc']/kwargs[0]['Rs'], tau, 3)

    def test_lenstronomy_params(self):

        m = 10 ** 8
        x = 0.5
        y = 1.0
        r3d = 100
        unique_tag = 1.0
        is_subhalo = False
        nfw_field_halo = TNFWFieldHalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, {},
                                      self.truncation_class, self.concentration_class, unique_tag)

        kwargs_halo, _ = nfw_field_halo.lenstronomy_params
        id = nfw_field_halo.lenstronomy_ID
        npt.assert_string_equal('TNFW', id[0])

        rho0, Rs, c, r200, M200 = self.lclenstronomy.nfw_angle2physical(kwargs_halo[0]['Rs'],
                                                                        kwargs_halo[0]['alpha_Rs'])
        npt.assert_almost_equal(M200/m, 1.0, 3)

        is_subhalo = True
        kwargs_profile = {'evaluate_mc_at_zlens': False}
        nfw_subhalo = TNFWSubhalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, kwargs_profile,
                                  self.truncation_class, self.concentration_class, unique_tag)
        c_subhalo = nfw_subhalo.c
        c_field = nfw_field_halo.c
        npt.assert_equal(True, c_subhalo <= c_field)

    def test_z_infall(self):
        m = 10 ** 8
        x = 0.5
        y = 1.0
        r3d = 100
        unique_tag = 1.0
        kwargs_profile = {'evaluate_mc_at_zlens': False, 'c_scatter': False, 'c_scatter_dex': 0.2}
        is_subhalo = True
        tnfw_subhalo = TNFWSubhalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, kwargs_profile,
                                  self.truncation_class, self.concentration_class, unique_tag)
        z_infall = tnfw_subhalo.z_infall
        npt.assert_equal(True, self.zhalo <= z_infall)

    def test_bound_mass(self):

        m = 10 ** 8
        x = 0.5
        y = 1.0
        r3d = 100
        unique_tag = 1.0
        kwargs_profile = {'evaluate_mc_at_zlens': False, 'c_scatter': False, 'c_scatter_dex': 0.2}
        is_subhalo = True
        truncation_class = TruncationRoche(None, 2.0)
        tnfw_subhalo = TNFWSubhalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, kwargs_profile,
                                  truncation_class, self.concentration_class, unique_tag)
        bound_mass = tnfw_subhalo.bound_mass
        npt.assert_equal(True, bound_mass/10**8 < 1)

    def test_density_mass(self):

        m = 10 ** 8
        x = 0.5
        y = 1.0
        r3d = 100
        unique_tag = 1.0
        kwargs_profile = {'evaluate_mc_at_zlens': False, 'c_scatter': False, 'c_scatter_dex': 0.2}
        is_subhalo = False
        tnfw_fieldhalo = TNFWFieldHalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, kwargs_profile,
                                  self.truncation_class, self.concentration_class, unique_tag)
        rho_s, rs, r200 = tnfw_fieldhalo.nfw_params
        c = tnfw_fieldhalo.c
        mtheory = 4 * np.pi * rho_s * rs ** 3 * (np.log(1 + c) - c/(1+c))
        npt.assert_almost_equal(mtheory, tnfw_fieldhalo.mass)
        rmax = c * rs
        m_calculated = tnfw_fieldhalo.mass_3d(rmax)
        npt.assert_almost_equal(mtheory/m_calculated, 1.0)

        tau = 2.
        tnfw_halo = TNFWFieldHalo.simple_setup(m, 0.0, 0.0, 0.6, tau, self.lens_cosmo)
        rs = tnfw_halo.nfw_params[1]
        r = np.linspace(0.001, tnfw_halo.c, 10000) * rs
        rho = tnfw_halo.density_profile_3d_lenstronomy(r)
        m_exact = np.trapz(4 * np.pi * rho * r ** 2, r)
        m_class = tnfw_halo.mass_3d('r200')
        npt.assert_almost_equal(m_class/ m_exact,1,3)

        tau = 1e-3
        tnfw_halo = TNFWFieldHalo.simple_setup(m, 0.0, 0.0, 0.6, tau, self.lens_cosmo)
        rs = tnfw_halo.nfw_params[1]
        r = np.linspace(0.0001, tnfw_halo.c, 100000) * rs
        rho = tnfw_halo.density_profile_3d_lenstronomy(r)
        m_exact = np.trapz(4 * np.pi * rho * r ** 2, r)
        m_class = tnfw_halo.mass_3d('r200')
        npt.assert_almost_equal(m_class / m_exact, 1, 3)

    def test_vmax(self):

        m = 10 ** 8
        x = 0.5
        y = 1.0
        r3d = 100
        unique_tag = 1.0
        kwargs_profile = {'evaluate_mc_at_zlens': False, 'c_scatter': False, 'c_scatter_dex': 0.2}
        is_subhalo = False
        tnfw_fieldhalo = TNFWFieldHalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, kwargs_profile,
                                       self.truncation_class, self.concentration_class, unique_tag)
        _, rs, _ = tnfw_fieldhalo.nfw_params
        r = np.linspace(0.01, 20, 5000) * rs
        m_enclosed = tnfw_fieldhalo.mass_3d(r)
        G = 4.3e-6
        v_circ = np.sqrt(G * m_enclosed / r)
        vmax_calc = max(v_circ)
        vmax = tnfw_fieldhalo.vmax_nfw
        npt.assert_almost_equal(vmax / vmax_calc, 1, 2)

        tnfw_subhalo = TNFWSubhalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, kwargs_profile,
                                       self.truncation_class, self.concentration_class, unique_tag)
        tnfw_subhalo._rescale_norm = 0.1
        _, rs, _ = tnfw_subhalo.nfw_params
        r = np.linspace(0.01, 20, 5000) * rs
        m_enclosed = tnfw_subhalo.mass_3d(r)
        G = 4.3e-6
        v_circ = np.sqrt(G * m_enclosed / r)
        vmax_calc = max(v_circ)
        vmax = tnfw_fieldhalo.vmax_nfw
        npt.assert_almost_equal(vmax / vmax_calc, np.sqrt(10), 1)

if __name__ == '__main__':
    pytest.main()




from pyHalo.single_realization import SingleHalo
from pyHalo.Halos.lens_cosmo import LensCosmo
from pyHalo.concentration_models import preset_concentration_models
from pyHalo.realization_extensions import RealizationExtensions
import numpy as np
from scipy.integrate import quad
import numpy.testing as npt
from copy import deepcopy
import pytest


class TestPromptCusp(object):

    def setup_method(self):

        halo_mass = 10 ** 7.7
        self.nfw_mass = halo_mass
        x_arcsec = 0.
        y_arcsec = 0.0
        z_halo = 0.5

        zsource = 1.5
        # the source redshift is needed to specify the center of the rendering volume, something necessary for full
        # halo populations. The source redshift won't actually matter for single-halo models. However, if you want to do
        # multi-plane lensing computations with lenstronomy, you will need the source redshift
        zlens = 0.5  # possibly irrelevant for single-halo models, depending on whether the object is a subhalo
        mass_definition = 'NFW'
        subhalo_flag = False
        lens_cosmo = LensCosmo(zlens, zsource)
        astropy_class = lens_cosmo.cosmo

        model, kwargs_concentration_model = preset_concentration_models('DIEMERJOYCE19')
        kwargs_concentration_model['scatter'] = False
        kwargs_concentration_model['cosmo'] = astropy_class
        concentration_model = model(**kwargs_concentration_model)

        halo_truncation_model = None  # an NFW halo has no formal truncation radius, so we don't need to specify this
        kwargs_halo_model = {'truncation_model': halo_truncation_model,
                             'concentration_model': concentration_model,
                             'kwargs_density_profile': {}}
        self.single_halo_realization = SingleHalo(halo_mass, x_arcsec, y_arcsec, mass_definition, z_halo, zlens, zsource,
                                 subhalo_flag, kwargs_halo_model=kwargs_halo_model)

    def test_rescaling_normalization(self):

        ext = RealizationExtensions(deepcopy(self.single_halo_realization))
        a = 0.1
        b = 0.0
        c = 0.0
        realization_with_PS = ext.add_prompt_cusps(a,b,c)
        nfw_halo = realization_with_PS.halos[0]
        ps_halo = realization_with_PS.halos[1]
        _ = nfw_halo.lenstronomy_params
        _ = ps_halo.lenstronomy_params
        npt.assert_almost_equal(nfw_halo._rescale_norm, 0.9)
        npt.assert_almost_equal(ps_halo._rescale_norm, 1.0)

    def test_mass_conservation(self):

        ext = RealizationExtensions(deepcopy(self.single_halo_realization))
        a = 0.25
        b = 0.0
        c = 0.0
        realization_with_PS = ext.add_prompt_cusps(a, b, c)
        nfw_halo = realization_with_PS.halos[0]
        ps_halo = realization_with_PS.halos[1]
        rescale_norm_nfw = nfw_halo._rescale_norm
        rs = nfw_halo.nfw_params[1]
        r = np.logspace(-1.5, 0.7, 100) * rs
        integrand_total = lambda x: 4 * np.pi * x ** 2 * (rescale_norm_nfw * nfw_halo.density_profile_3d(x) +
                                                          ps_halo.density_profile_3d(x))
        integrand_ps = lambda x: 4 * np.pi * x ** 2 * ps_halo.density_profile_3d(x)

        # test that the total mass of the ps_halo is what it should be
        mass_target = nfw_halo.mass * a
        _, cusp_R, cusp_A = ps_halo.profile_args
        ps_halo_mass = 8 * np.pi / 3 * cusp_A * cusp_R ** 1.5
        npt.assert_almost_equal(np.log10(mass_target), np.log10(ps_halo_mass))
        npt.assert_almost_equal(np.log10(mass_target) / np.log10(quad(integrand_ps, 0, nfw_halo.c * rs))[0], 1, 2)


        total_mass = quad(integrand_total, 0, nfw_halo.c * rs)[0]
        npt.assert_array_less(abs(100*(total_mass / nfw_halo.mass - 1)), 1.6)

if __name__ == '__main__':
    pytest.main()



from pyHalo.Halos.halo_base import Halo
from lenstronomy.LensModel.Profiles.nfw_core_truncated import TNFWC as TNFWCLenstronomy
import numpy as np
from copy import deepcopy

_tnfwc_lenstronomy = TNFWCLenstronomy()


class TNFWCHaloParametric(Halo):
    """
    The same profile as TNFWCHalo, but gets initialized with the core radius, truncation radius, and scale radius as
    profile parameters, rather than an SIDM timescale as a parameter
    """

    """
    The base class for a cored and truncated NFW halo
    """
    _pseudo_nfw = False # for the parametric profile class this should be False

    def __init__(self, mass, x, y, r3d, z,
                 sub_flag, lens_cosmo_instance, args,
                 truncation_class, concentration_class, unique_tag):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        self._lens_cosmo = lens_cosmo_instance
        self._concentration_class = concentration_class
        self._truncation_class = truncation_class
        self.tnfwc_lenstronomy = TNFWCLenstronomy()
        mdef = 'TNFWC'
        super(TNFWCHaloParametric, self).__init__(mass, x, y, r3d, mdef, z, sub_flag,
                                        lens_cosmo_instance, args, unique_tag)

    def density_profile_3d(self, r, profile_args=None):
        """

        :param r: distance from center of halo [kpc]
        :param profile_args:
        :return:
        """
        return self.density_profile_3d_lenstronomy(r)

    def density_profile_3d_lenstronomy(self, r, kwargs_lenstronomy=None):
        """
        Computes the 3-D density profile of the halo
        :param r: distance from center of halo [kpc]
        :return: the density profile in units M_sun / kpc^3
        """
        if kwargs_lenstronomy is None:
            kwargs_lenstronomy = self.lenstronomy_params[0][0]
        kpc_per_arcsec = self._lens_cosmo.cosmo.kpc_proper_per_asec(self.z)
        sigma_crit_mpc = self._lens_cosmo.get_sigma_crit_lensing(self.z, self._lens_cosmo.z_source)
        sigma_crit_kpc = sigma_crit_mpc * 1e-6
        factor = sigma_crit_kpc / kpc_per_arcsec
        return factor * self.tnfwc_lenstronomy.density_lens(r / kpc_per_arcsec,
                                                            kwargs_lenstronomy['Rs'],
                                                            kwargs_lenstronomy['alpha_Rs'],
                                                            kwargs_lenstronomy['r_core'],
                                                            kwargs_lenstronomy['r_trunc'])

    @property
    def c(self):
        """
        Computes the halo concentration (once)
        """
        if not hasattr(self, '_c'):
            self._c = self._concentration_class.nfw_concentration(self.mass, self.z_eval)
        return self._c

    @property
    def lenstronomy_ID(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """

        return ['TNFWC']

    @property
    def lenstronomy_params(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_kwargs_lenstronomy'):
            [alpha_Rs, rs_kpc, rc_kpc, rt_kpc, _] = self.profile_args
            kpc_per_arcsec = self._lens_cosmo.cosmo.kpc_proper_per_asec(self.z)
            Rs_angle = rs_kpc / kpc_per_arcsec
            r_core_angle = rc_kpc / kpc_per_arcsec
            r_trunc_angle = rt_kpc / kpc_per_arcsec
            x, y = np.round(self.x, 4), np.round(self.y, 4)
            kwargs = [{'alpha_Rs': alpha_Rs,
                       'Rs': Rs_angle,
                       'center_x': x, 'center_y': y,
                       'r_trunc': r_trunc_angle,
                       'r_core': r_core_angle}]
            self._kwargs_lenstronomy = kwargs
        return self._kwargs_lenstronomy, None

    @property
    def profile_args(self):
        """
        Computes the time-evolving profile parameters
        :return: the density normalization, scale radius, and core radius for thee SIDM halo
        """
        if not hasattr(self, '_profile_args'):
            _, rs_kpc0, _ = self.lens_cosmo.NFW_params_physical(self.mass, self.c, self.z_eval,
                                                               pseudo_nfw=False)
            r200_kpc = rs_kpc0 * self.c
            r_match_kpc = self.c * rs_kpc0
            kpc_per_arcsec = self._lens_cosmo.cosmo.kpc_proper_per_asec(self.z)
            rt_kpc = self._args['rt_kpc']
            r_trunc_angle = rt_kpc / kpc_per_arcsec
            if rt_kpc < self._args['x_core_halo'] * rs_kpc0:
                rs_kpc = rt_kpc * 0.99
            else:
                rs_kpc = self._args['x_core_halo'] * rs_kpc0
            r_core_kpc = rs_kpc * 0.99
            x = np.logspace(-4,
                            np.log10(r_match_kpc / rs_kpc0),
                            1000)
            r = x * rs_kpc0
            rs_angle = rs_kpc / kpc_per_arcsec
            r_core_angle = r_core_kpc / kpc_per_arcsec
            kwargs_temp = {'alpha_Rs': 1.0,
                           'Rs': rs_angle,
                           'r_core': r_core_angle,
                           'r_trunc': r_trunc_angle}
            rho = self.density_profile_3d_lenstronomy(r, kwargs_temp)
            mass_3d = np.trapz(4 * np.pi * r ** 2 * rho, r)
            alpha_Rs = self._args['mass_conservation'] / mass_3d
            self._profile_args = (alpha_Rs,
                                  rs_kpc,
                                  r_core_kpc,
                                  rt_kpc,
                                  r200_kpc)
        return self._profile_args

    @property
    def params_physical(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_params_physical'):
            [rho_s, rs_kpc, rc_kpc, rt_kpc, r200_kpc] = self.profile_args
            self._params_physical = {'rhos': rho_s * self._rescale_norm,
                                     'rs': rs_kpc, 'r200': r200_kpc,
                                     'r_trunc_kpc': rt_kpc,
                                     'r_core_kpc': rc_kpc}
        return self._params_physical

    @property
    def vmax_nfw(self):
        """
        Returns the maximum circular velocity in km/sec of an NFW profile with given rhos, rs
        :return:
        """
        if not hasattr(self, '_vmax'):
            rhos, rs, _ = self.nfw_params
            _ = self.profile_args
            self._vmax = self._lens_cosmo.nfw_vmax(self._rescale_norm * rhos, rs)
        return self._vmax


class TNFWCHaloEvolving(TNFWCHaloParametric):
    """
    The base class for a cored and truncated NFW halo
    """
    _pseudo_nfw = True

    @property
    def halo_effective_age(self):
        if not hasattr(self, '_halo_effective_age'):
            if self.is_subhalo:
                self._halo_effective_age = self.lens_cosmo.sidm_halo_effective_age(self.z,
                                                                               self.z_infall,
                                                                               self._args['lambda_t'])
            else:
                self._halo_effective_age = self.halo_age
        return self._halo_effective_age

    @property
    def sidm_timescale(self):
        """
        Computes the timescale given by Equation 2.2 in https://arxiv.org/pdf/2305.16176.pdf
        :return:
        """
        return self._args['sidm_timescale']

    @property
    def t_over_tc(self):
        """
        Computes the dimensionless timescale for the halo evolution
        :return:
        """
        return self.halo_effective_age / self.sidm_timescale

    @property
    def profile_args(self):
        """
        Computes the time-evolving profile parameters
        :return: the density normalization, scale radius, and core radius for thee SIDM halo
        """
        if not hasattr(self, '_profile_args'):
            rt_kpc = self._args['rt_kpc']
            _, rs_0, _ = self.lens_cosmo.NFW_params_physical(self.mass, self.c, self.z_eval,
                                                             pseudo_nfw=self._pseudo_nfw)
            rs_kpc, rc_kpc = evolve_profile(self.t_over_tc, rs_0)
            kpc_per_arcsec = self._lens_cosmo.cosmo.kpc_proper_per_asec(self.z)
            r200_kpc = self.c * rs_0
            # now solve for rho_s to conserve total mass
            Rs_angle = rs_kpc / kpc_per_arcsec  # Rs in arcsec
            r_core_angle = rc_kpc / kpc_per_arcsec
            r_trunc_angle = rt_kpc / kpc_per_arcsec
            x = np.logspace(-4,
                            np.log10(self.c),
                            1000)
            r = x * rs_0
            kwargs_temp = {'alpha_Rs': 1.0,
                           'Rs': Rs_angle,
                           'r_core': r_core_angle,
                           'r_trunc': r_trunc_angle}
            rho = self.density_profile_3d_lenstronomy(r, kwargs_temp)
            mass_3d = np.trapz(4 * np.pi * r ** 2 * rho, r)
            alpha_Rs = self._args['mass_conservation'] / mass_3d
            self._profile_args = (alpha_Rs, rs_kpc, rc_kpc, rt_kpc, r200_kpc)
        return self._profile_args


class Hybrid(Halo):

    """
    A hybrid TNFW + TNFWC profile with a relative weighting between them
    """
    def __init__(self, tnfw_halo, tnfwc_halo, rescaling_factor):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        c = tnfw_halo.c
        tnfwc_halo._c = c
        if tnfw_halo.is_subhalo:
            assert tnfwc_halo.is_subhalo
            tnfwc_halo._z_infall = tnfw_halo.z_infall
        self.tnfw_halo = tnfw_halo
        self.tnfwc_halo = tnfwc_halo
        if hasattr(self.tnfw_halo, '_kwargs_lenstronomy'):
            delattr(self.tnfw_halo, '_kwargs_lenstronomy')
        if hasattr(self.tnfwc_halo, '_kwargs_lenstronomy'):
            delattr(self.tnfwc_halo, '_kwargs_lenstronomy')
        mdef = 'TNFWC_HYBRID'
        self._rescaling_factor = rescaling_factor
        super(Hybrid, self).__init__(self.tnfw_halo.mass, self.tnfw_halo.x, self.tnfw_halo.y,
                                     self.tnfw_halo.r3d, mdef, self.tnfw_halo.z,
                                     self.tnfw_halo.is_subhalo, self.tnfw_halo.lens_cosmo,
                                     self.tnfw_halo._args, self.tnfw_halo.unique_tag)

    @property
    def vmax_nfw(self):
        """
        Returns the maximum circular velocity in km/sec of an NFW profile with given rhos, rs
        :return:
        """
        return self.tnfw_halo.vmax_nfw

    @property
    def halo_effective_age(self):
        return self.tnfwc_halo.halo_effective_age

    @property
    def t_over_tc(self):
        """

        :return:
        """
        return self.tnfwc_halo.t_over_tc

    @property
    def sidm_timescale(self):
        """

        :return:
        """
        return self.tnfwc_halo.sidm_timescale

    @property
    def c(self):
        """
        Computes the halo concentration (once)
        """
        if not hasattr(self, '_c'):
            self._c = self.tnfw_halo.c
        return self._c

    def profile_args(self):
        return [self.tnfw_halo.profile_args(), self.tnfwc_halo.profile_args()]

    @property
    def lenstronomy_params(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_kwargs_lenstronomy'):
            lenstronomy_params_tnfw = deepcopy(self.tnfw_halo.lenstronomy_params[0])
            lenstronomy_params_tnfwc = deepcopy(self.tnfwc_halo.lenstronomy_params[0])
            lenstronomy_params_tnfw[0]['alpha_Rs'] *= 1-self._rescaling_factor
            lenstronomy_params_tnfwc[0]['alpha_Rs'] *= self._rescaling_factor
            self._kwargs_lenstronomy = lenstronomy_params_tnfw + lenstronomy_params_tnfwc
        return self._kwargs_lenstronomy, None

    @property
    def nfw_params(self):
        """
        Computes the nfw profile parameters (rs,rho_s) from mass and concentration
        :return: rs, r200 and rho_s in units kpc, kpc, and M_sun / kpc^3
        """
        if not hasattr(self, '_nfw_params'):
            self._nfw_params_tnfw = self.tnfw_halo.nfw_params
            self._nfw_params_tnfwc = self.tnfwc_halo.nfw_params
            self._nfw_params = [self._nfw_params_tnfw, self._nfw_params_tnfwc]
        return [self._nfw_params[0][0], self._nfw_params[1][0]], \
               [self._nfw_params[0][1], self._nfw_params[1][1]], \
               [self._nfw_params[0][2], self._nfw_params[1][2]]

    @property
    def lenstronomy_ID(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        return ['TNFW', 'TNFWC']

    def density_profile_3d_lenstronomy(self, r, profile_args=None):
        """
        Computes the 3-D density profile of the halo
        :param r: distance from center of halo [kpc]
        :return: the density profile in units M_sun / kpc^3
        """
        rho_tnfw = self.tnfw_halo.density_profile_3d_lenstronomy(r)
        rho_tnfwc = self.tnfwc_halo.density_profile_3d_lenstronomy(r)
        return rho_tnfw * (1-self._rescaling_factor) + rho_tnfwc * self._rescaling_factor

    def density_profile_3d(self, r, profile_args=None):
        """
        Computes the 3-D density profile of the halo
        :param r: distance from center of halo [kpc]
        :return: the density profile in units M_sun / kpc^3
        """
        rho_tnfw = self.tnfw_halo.density_profile_3d(r)
        rho_tnfwc = self.tnfwc_halo.density_profile_3d(r)
        return rho_tnfw * (1-self._rescaling_factor) + rho_tnfwc * self._rescaling_factor

def evolve_profile(t, rs_0):

    t = min(1.6, t)
    rs = rs_0 * rs_evolution(t)
    rc = rs_0 * rc_evolution(t)
    rc = max(1e-7 * rs, rc)
    return rs, rc

def rs_evolution(tr):
    if tr <= 1.0:
        return _rs_evolution(tr)
    else:
        return rs_extrapolation(tr)


def rc_evolution(tr):
    if tr <= 1.0:
        return _rc_evolution(tr)
    else:
        return rc_extrapolation(tr)


def rho_s_evolution(tr):
    """
    Computes the evolution of the density normalization for the profile
    :param t_over_tc: the physical time scaled by the characteristic evolution time
    :return: the density in units of rho_s_nfw
    """
    return 0.692896 + 1.11215 * tr + 4.08729 * tr ** 5 - 5.78038 * tr ** 7 + 5.30498 * tr ** 9 + (
        1 - 0.692896) / np.log(0.001) * np.log(0.001 + tr)


def _rs_evolution(tr):
    """
    Computes the evolution of the scale radius for the profile
    :param t_over_tc: the physical time scaled by the characteristic evolution time
    :return: the scale radius in units of rs_nfw
    """

    return 1.13293 - 0.518186 * tr + 0.30265 * tr ** 2 - 0.400463 * tr ** 3 + (1 - 1.13293) * np.log(
        0.001 + tr) / np.log(0.001)


def _rc_evolution(tr):
    """
    Computes the evolution of the core radius for the profile
    :param t_over_tc: the physical time scaled by the characteristic evolution time
    :return: the core radius in units of rs_nfw
    """

    return 1.06419 * np.sqrt(tr) - 1.33207 * tr + 0.431133 * tr ** 2 - 0.148087 * tr ** 3 + 0.024455 * tr ** 4


def rs_extrapolation(tr, t0=0.5, c1=0.47, c2=2.4):
    """

    :param tr:
    :return:
    """
    return rs_evolution(t0) * np.exp(-c1 * (tr - t0) - c2 * (tr - t0) ** 3)


def rc_extrapolation(t_over_tc, t0=0.5, c1=0.28):
    """

    :param t_over_tc:
    :param t0:
    :param c1:
    :return:
    """
    rc = rc_evolution(t0) - c1 * (t_over_tc - t0)
    if rc < 0.0001:
        rc = 0.0001
    return rc



import numpy.testing as npt
import numpy as np
from lenstronomy.LensModel.Profiles.nfw_core_truncated import TNFWC as TNFWCLenstronomy
from lenstronomy.LensModel.Profiles.nfw import NFW as NFWLenstronomy
from pyHalo.Halos.HaloModels.NFW_core_trunc import TNFWCHaloEvolving
from pyHalo.truncation_models import ConstantTruncationArcsec
from pyHalo.concentration_models import ConcentrationConstant
from pyHalo.Halos.lens_cosmo import LensCosmo
import pytest


class TestTNFWC(object):

    def setup_method(self):

        self.lens_cosmo = LensCosmo(0.5, 2.5)
        self.tnfwc_lenstronomy = TNFWCLenstronomy()
        self.nfw_lenstronomy = NFWLenstronomy()
        self.truncation_class = ConstantTruncationArcsec(self.lens_cosmo, 1000.0)
        self.concentration_class = ConcentrationConstant(None, 10.0)

    def test_mass_conservation(self):

        mass0 = 10 ** 8
        kwargs_profile = {'sidm_timescale': 20.,
                          'lambda_t': 1.0,
                          'mass_conservation': mass0,
                          'rt_kpc': 1000}
        tnfwc = TNFWCHaloEvolving(mass0, 0.0, 0.0, None, 0.5, False,
                               self.lens_cosmo, kwargs_profile, self.truncation_class,
                               self.concentration_class, 1.0)
        r = np.linspace(0.001, tnfwc.c, 10000) * tnfwc.nfw_params[1]
        mass = np.trapz(4*np.pi*r**2*tnfwc.density_profile_3d_lenstronomy(r),r)

        npt.assert_almost_equal(mass / mass0, 1, 2)

        kwargs_profile = {'sidm_timescale': 10.,
                          'lambda_t': 1.0,
                          'mass_conservation': mass0,
                          'rt_kpc': 1000}
        tnfwc = TNFWCHaloEvolving(mass0, 0.0, 0.0, None, 0.5, False,
                          self.lens_cosmo, kwargs_profile, self.truncation_class,
                          self.concentration_class, 1.0)
        mass = np.trapz(4*np.pi*r**2*tnfwc.density_profile_3d_lenstronomy(r),r)
        npt.assert_almost_equal(mass / mass0, 1, 2)

        kwargs_profile = {'sidm_timescale': 2.5,
                          'lambda_t': 1.0,
                          'mass_conservation': mass0,
                          'rt_kpc': 1000}
        tnfwc = TNFWCHaloEvolving(mass0, 0.0, 0.0, None, 0.5, False,
                          self.lens_cosmo, kwargs_profile, self.truncation_class,
                          self.concentration_class, 1.0)
        mass = np.trapz(4*np.pi*r**2*tnfwc.density_profile_3d_lenstronomy(r),r)
        npt.assert_array_less(abs(mass / mass0 - 1), 0.016)

        kwargs_profile = {'sidm_timescale': 0.5,
                          'lambda_t': 1.0,
                          'mass_conservation': mass0,
                          'rt_kpc': 1000}
        tnfwc = TNFWCHaloEvolving(mass0, 0.0, 0.0, None, 0.5, False,
                          self.lens_cosmo, kwargs_profile, self.truncation_class,
                          self.concentration_class, 1.0)
        mass = np.trapz(4*np.pi*r**2*tnfwc.density_profile_3d_lenstronomy(r),r)
        npt.assert_array_less(abs(mass / mass0 - 1), 0.016)

    def test_vmax(self):
        """Here we test that these objects have the vmax of the reference TNFW/NFW halo profile"""

        mass0 = 10 ** 8
        kwargs_profile = {'sidm_timescale': 0.5,
                          'lambda_t': 1.0,
                          'mass_conservation': mass0,
                          'rt_kpc': 1000}
        tnfwc = TNFWCHaloEvolving(mass0, 0.0, 0.0, None, 0.5, False,
                          self.lens_cosmo, kwargs_profile, self.truncation_class,
                          self.concentration_class, 1.0)
        vmax = tnfwc.vmax_nfw
        rhos, rs, _ = tnfwc.nfw_params
        vmax_calc = self.lens_cosmo.nfw_vmax(rhos, rs)
        npt.assert_almost_equal(vmax / vmax_calc, 1, 2)

if __name__ == '__main__':
    pytest.main()



import numpy.testing as npt
from pyHalo.Halos.HaloModels.sis import SIS, MassiveGalaxy
from pyHalo.Halos.HaloModels.TNFW import TNFWFieldHalo, TNFWSubhalo
from pyHalo.Halos.concentration import ConcentrationDiemerJoyce
from pyHalo.Halos.tidal_truncation import TruncationRoche, Multiple_RS
from astropy.cosmology import FlatLambdaCDM
from pyHalo.Halos.lens_cosmo import LensCosmo
from pyHalo.Cosmology.cosmology import Cosmology
from lenstronomy.LensModel.Profiles.sis import SIS as SISLenstronomy
from lenstronomy.LensModel.Profiles.pseudo_double_powerlaw import PseudoDoublePowerlaw
import pytest
import numpy as np


class TestMassiveGalaxy(object):

    def setup_method(self):
        astropy = FlatLambdaCDM(70.0, 0.3)
        cosmo = Cosmology(astropy_instance=astropy)
        self.zhalo = 0.5
        self.zsource = 2.0
        self.lens_cosmo = LensCosmo(self.zhalo, self.zsource, cosmo)
        self.truncation_class = TruncationRoche(None, 100000000.0)
        self.concentration_class = ConcentrationDiemerJoyce(self.lens_cosmo, scatter=False)

    def test_lenstronomy_params(self):

        m = 10 ** 11.3
        x = 0.5
        y = 1.0
        r3d = 100
        unique_tag = 1.0
        is_subhalo = False
        nfw_field_halo = TNFWFieldHalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, {},
                                       self.truncation_class, self.concentration_class, unique_tag)

        sis = SIS(nfw_field_halo)
        vdis = sis.profile_args
        D_ds = nfw_field_halo.lens_cosmo.cosmo.D_A(0.5, 2.0)
        D_s = nfw_field_halo.lens_cosmo.cosmo.D_A(0., 2.0)
        arcsec = 206265
        c = 299792.5  # in km/sec
        thetaE = 4*np.pi*vdis**2/c**2 * D_ds / D_s * arcsec
        npt.assert_almost_equal(thetaE, sis.lenstronomy_params[0][0]['theta_E'])

    def test_lenstronomy_ID(self):

        m = 10 ** 11.3
        x = 0.5
        y = 1.0
        r3d = 100
        unique_tag = 1.0
        is_subhalo = False
        nfw_field_halo = TNFWFieldHalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, {},
                                       self.truncation_class, self.concentration_class, unique_tag)

        sis = SIS(nfw_field_halo)
        npt.assert_string_equal('SIS', sis.lenstronomy_ID[0])

    def test_sis_mass(self):

        m = 10 ** 11.3
        x = 0.5
        y = 1.0
        r3d = 100
        unique_tag = 1.0
        is_subhalo = False
        nfw_field_halo = TNFWFieldHalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, {},
                                       self.truncation_class, self.concentration_class, unique_tag)

        sis = SIS(nfw_field_halo)
        thetaE = sis.lenstronomy_params[0][0]['theta_E']
        _, rs, r200 = nfw_field_halo.nfw_params

        sis_lenstronomy = SISLenstronomy()
        mass3d = sis_lenstronomy.mass_3d_lens(rs, thetaE) * nfw_field_halo.lens_cosmo.sigmacrit
        npt.assert_almost_equal(mass3d/nfw_field_halo.mass, 0.857, 2)

    def test_gnfw_mass(self):

        m = 10 ** 11.3
        x = 0.5
        y = 1.0
        r3d = 100
        unique_tag = 1.0
        is_subhalo = False
        nfw_field_halo = TNFWFieldHalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, {},
                                       self.truncation_class, self.concentration_class, unique_tag)
        gal = MassiveGalaxy(nfw_field_halo)
        kwargs_lenstronomy = gal.lenstronomy_params[0][0]
        prof = PseudoDoublePowerlaw()
        kpc_per_arcsec = nfw_field_halo.lens_cosmo.cosmo.kpc_proper_per_asec(gal.z)
        r = nfw_field_halo.nfw_params[-1]/kpc_per_arcsec
        sigma_crit_mpc = nfw_field_halo.lens_cosmo.get_sigma_crit_lensing(gal.z, nfw_field_halo.lens_cosmo.z_source)
        sigma_crit_arcsec = sigma_crit_mpc * (0.001 * kpc_per_arcsec) ** 2
        mass3d = sigma_crit_arcsec*prof.mass_3d_lens(r,
                            kwargs_lenstronomy['Rs'],
                           kwargs_lenstronomy['alpha_Rs'],
                           kwargs_lenstronomy['gamma_inner'],
                           kwargs_lenstronomy['gamma_outer'])
        npt.assert_almost_equal(mass3d/nfw_field_halo.mass, 1.0, 2)

        m = 10 ** 11.7
        x = 0.
        y = 0.0
        r3d = 100
        unique_tag = 1.0
        is_subhalo = True
        tau = 2
        truncation_class = Multiple_RS(self.lens_cosmo, tau)
        nfw_sub_halo = TNFWSubhalo(m, x, y, r3d, self.zhalo, is_subhalo, self.lens_cosmo, {},
                                       truncation_class, self.concentration_class, unique_tag)
        gal = MassiveGalaxy(nfw_sub_halo)
        kwargs_lenstronomy = gal.lenstronomy_params[0][0]
        prof = PseudoDoublePowerlaw()
        kpc_per_arcsec = nfw_field_halo.lens_cosmo.cosmo.kpc_proper_per_asec(gal.z)
        r = nfw_sub_halo.nfw_params[-1] / kpc_per_arcsec
        sigma_crit_mpc = nfw_field_halo.lens_cosmo.get_sigma_crit_lensing(gal.z, nfw_field_halo.lens_cosmo.z_source)
        sigma_crit_arcsec = sigma_crit_mpc * (0.001 * kpc_per_arcsec) ** 2
        npt.assert_almost_equal(kwargs_lenstronomy['Rs'], nfw_sub_halo.profile_args[1]/kpc_per_arcsec, 3)

        mass3d = sigma_crit_arcsec * prof.mass_3d_lens(r,
                                                       kwargs_lenstronomy['Rs'],
                                                       kwargs_lenstronomy['alpha_Rs'],
                                                       kwargs_lenstronomy['gamma_inner'],
                                                       kwargs_lenstronomy['gamma_outer'])
        npt.assert_almost_equal(mass3d / nfw_sub_halo.bound_mass, 1.0, 5)
        #
        # import matplotlib.pyplot as plt
        # rs = nfw_sub_halo.nfw_params[1]
        # r_plot = np.logspace(-2, 1., 1000) * rs
        # rho_nfw = nfw_sub_halo.density_profile_3d(r_plot)
        # rho_gal = gal.density_profile_3d_lenstronomy(r_plot)
        # plt.loglog(r_plot, rho_nfw)
        # plt.loglog(r_plot, rho_gal)
        # #print(rho_nfw)
        # plt.show()

if __name__ == '__main__':
    pytest.main()




from pyHalo.Halos.tidal_truncation import TruncationRN, TruncationRoche, \
    TruncationSplashBack, TruncateMeanDensity, \
    TruncationGalacticus, ConstantTruncationArcsec, TruncationGalacticusKeeley24, Multiple_RS, TruncationBoundMassPDF


def truncation_models(model_name):
    """
    Loads and returns methods to set the tidal truncation radius (or equivalent parameter) of halos
    """
    kwargs_model = {}
    if isinstance(model_name, str):
        if model_name == 'TRUNCATION_R50':
            kwargs_model['LOS_truncation_factor'] = 50
            # truncates NFW halos at r50 (for field halos, this is comparable to the splashback radius)
            return TruncationRN, kwargs_model
        elif model_name == 'TRUNCATION_R200':
            kwargs_model['LOS_truncation_factor'] = 200
            # truncates NFW halos at r50 (for field halos, this is comparable to the splashback radius)
            return TruncationRN, kwargs_model
        elif model_name == 'TRUNCATION_RN':
            return TruncationRN, kwargs_model
        elif model_name == 'TRUNCATION_ROCHE':
            return TruncationRoche, kwargs_model
        elif model_name == 'TRUNCATION_ROCHE_GILMAN2020':
            kwargs_model['RocheNorm'] = 1.4
            kwargs_model['m_power'] = 1./3
            kwargs_model['RocheNu'] = 2./3
            return TruncationRoche, kwargs_model
        elif model_name == 'SPLASHBACK':
            return TruncationSplashBack, kwargs_model
        elif model_name == 'TRUNCATION_MEAN_DENSITY':
            return TruncateMeanDensity, kwargs_model
        elif model_name == 'TRUNCATION_GALACTICUS':
            return TruncationGalacticus, kwargs_model
        elif model_name == 'TRUNCATION_GALACTICUS_KEELEY24':
            return TruncationGalacticusKeeley24, kwargs_model
        elif model_name == 'CONSTANT':
            return ConstantTruncationArcsec, kwargs_model
        elif model_name == 'MULTIPLE_RS':
            return Multiple_RS, kwargs_model
        elif model_name == 'BOUND_MASS_PDF':
            return TruncationBoundMassPDF, kwargs_model
        else:
            raise Exception('model '+str(model_name)+' not recognized')
    else:
        # can also pass in a class directly
        return model_name, kwargs_model



from pyHalo.Halos.halo_base import Halo
import numpy as np
from lenstronomy.LensModel.Profiles.pseudo_double_powerlaw import PseudoDoublePowerlaw


class SIS(Halo):
    """
    The base class for an SIS object; this is intended to represent LMC-like objects with enough luminous material
    that they do not have NFW profiles

    This class is intended instantiated from an NFW halo class
    """
    def __init__(self, nfw_halo):
        """

        :param nfw_halo: an instance of TNFW, or TNFWCHalo halo classes
        """
        self._lens_cosmo = nfw_halo.lens_cosmo
        mdef = 'SIS'
        self._nfw_halo = nfw_halo
        super(SIS, self).__init__(mass=nfw_halo.mass,
                                  x=nfw_halo.x,
                                  y=nfw_halo.y,
                                  r3d=nfw_halo.r3d,
                                  mdef=mdef,
                                  z=nfw_halo.z,
                                  sub_flag=nfw_halo.is_subhalo,
                                  lens_cosmo_instance=nfw_halo.lens_cosmo,
                                  args=nfw_halo._args,
                                  unique_tag=nfw_halo.unique_tag,
                                  fixed_position=nfw_halo.fixed_position)

    @property
    def profile_args(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_profile_args'):
            vmax = self._nfw_halo.vmax_nfw
            velocity_dispersion = 0.646 * vmax
            self._profile_args = (velocity_dispersion)
        return self._profile_args

    @property
    def bound_mass(self):
        if self.is_subhalo:
            return self._nfw_halo.bound_mass
        else:
            raise Exception('field halos do not have bound mass attribute')

    @property
    def lenstronomy_ID(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        return ['SIS']

    @property
    def lenstronomy_params(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_kwargs_lenstronomy'):
            sigma_v = self.profile_args
            theta_E = self._lens_cosmo.thetaE_from_sigma(self.z, sigma_v)
            #mass3d = sis_lenstronomy.mass_3d_lens(rs, thetaE) * nfw_field_halo.lens_cosmo.sigmacrit
            x, y = np.round(self.x, 4), np.round(self.y, 4)
            kwargs = [{'theta_E': theta_E, 'center_x': x,  'center_y': y}]
            self._kwargs_lenstronomy = kwargs
        return self._kwargs_lenstronomy, None

class MassiveGalaxy(Halo):
    _pseudo_nfw = True
    def __init__(self, nfw_halo):
        """

        :param nfw_halo: an instance of TNFW, or TNFWCHalo halo classes
        """
        self._lens_cosmo = nfw_halo.lens_cosmo
        mdef = 'GNFW'
        self._nfw_halo = nfw_halo
        self._prof = PseudoDoublePowerlaw()
        halo_args = {'gamma_inner': 2.0, 'gamma_outer': 5.0, 'x_match': 'c'}
        super(MassiveGalaxy, self).__init__(mass=nfw_halo.mass,
                                  x=nfw_halo.x,
                                  y=nfw_halo.y,
                                  r3d=nfw_halo.r3d,
                                  mdef=mdef,
                                  z=nfw_halo.z,
                                  sub_flag=nfw_halo.is_subhalo,
                                  lens_cosmo_instance=nfw_halo.lens_cosmo,
                                  args=halo_args,
                                  unique_tag=nfw_halo.unique_tag,
                                  fixed_position=nfw_halo.fixed_position)

    def density_profile_3d_lenstronomy(self, r, profile_args=None):
        """
        Computes the 3-D density profile of the halo
        :param r: distance from center of halo [kpc]
        :return: the density profile in units M_sun / kpc^3
        """
        prof = self._prof
        kwargs_lenstronomy = self.lenstronomy_params[0][0]
        kpc_per_arcsec = self._lens_cosmo.cosmo.kpc_proper_per_asec(self.z)
        sigma_crit_mpc = self._lens_cosmo.get_sigma_crit_lensing(self.z, self._lens_cosmo.z_source)
        sigma_crit_kpc = sigma_crit_mpc * 1e-6
        factor = sigma_crit_kpc / kpc_per_arcsec
        return factor*prof.density_lens(r / kpc_per_arcsec,
                                        kwargs_lenstronomy['Rs'],
                                        kwargs_lenstronomy['alpha_Rs'],
                                        kwargs_lenstronomy['gamma_inner'],
                                        kwargs_lenstronomy['gamma_outer'])


    @property
    def bound_mass(self):
        """
        Return the bound mass (if is a subhalo)
        :return:
        """
        if self.is_subhalo:
            return self._nfw_halo.bound_mass
        else:
            raise Exception('field halos do not have bound mass attribute')

    @property
    def c(self):
        """
        Computes the halo concentration (once)
        """
        if not hasattr(self, '_c'):
            self._c = self._nfw_halo.c
        return self._c

    @property
    def lenstronomy_params(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_lenstronomy_args'):
            (concentration, gamma_inner, gamma_outer) = self.profile_args
            _, _, r200 = self.nfw_params
            kpc_per_arcsec = self._lens_cosmo.cosmo.kpc_proper_per_asec(self.z)
            if self._nfw_halo.is_subhalo:
                rs = self._nfw_halo.profile_args[1]
            else:
                rs = self._nfw_halo.nfw_params[1]
            rs_arcsec = rs / kpc_per_arcsec
            alpha_Rs = self._alphaRs()
            x, y = np.round(self.x, 4), np.round(self.y, 4)
            rs_arcsec = np.round(rs_arcsec, 10)
            alpha_Rs = np.round(alpha_Rs, 10)
            self._lenstronomy_args = [{'alpha_Rs': alpha_Rs, 'Rs': rs_arcsec,
                                       'gamma_inner': gamma_inner,
                                       'center_x': x, 'center_y': y,
                                      'gamma_outer': gamma_outer}]
        return self._lenstronomy_args, None

    @property
    def lenstronomy_ID(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        return ['PSEUDO_DPL']

    @property
    def profile_args(self):
        """
        See documentation in base class (Halos/halo_base.py)
        """
        if not hasattr(self, '_profile_args'):

            concentration = self._nfw_halo.c
            _ = self._nfw_halo.profile_args
            gamma_inner = self._args['gamma_inner']
            gamma_outer = self._args['gamma_outer']
            self._profile_args = (concentration, gamma_inner, gamma_outer)

        return self._profile_args

    def _alphaRs(self):
        """
        Calculates the density normalization that conserves mass within x_match
        :return:
        """
        if not hasattr(self, '_rho0_norm'):
            kpc_per_arcsec = self._lens_cosmo.cosmo.kpc_proper_per_asec(self.z)
            m_nfw = self._nfw_halo.mass_3d('r200')
            if self._nfw_halo.is_subhalo:
                rs_kpc = self._nfw_halo.profile_args[1]
            else:
                rs_kpc = self._nfw_halo.nfw_params[1]
            _, _, r200_kpc = self._nfw_halo.nfw_params
            rs_arcsec = rs_kpc / kpc_per_arcsec
            r_match_arcsec = r200_kpc / kpc_per_arcsec
            gamma_inner = self._args['gamma_inner']
            gamma_outer = self._args['gamma_outer']
            sigma_crit_mpc = self._lens_cosmo.get_sigma_crit_lensing(self.z, self._lens_cosmo.z_source)
            sigma_crit_arcsec = sigma_crit_mpc * (0.001 * kpc_per_arcsec) ** 2
            self._alphaRs = m_nfw / self._prof.mass_3d_lens(r_match_arcsec,
                                                            rs_arcsec,
                                                            1.0,
                                                            gamma_inner,
                                                            gamma_outer) / sigma_crit_arcsec
        return self._alphaRs








import numpy
from scipy.optimize import minimize
from lenstronomy.Cosmo.nfw_param import NFWParam
import astropy.units as un
from colossus.lss.bias import twoHaloTerm
from scipy.integrate import quad
from pyHalo.Halos.accretion import InfallDistributionGalacticus2024, InfallDistributionHybrid


class NFWParampyHalo(NFWParam):

    """Adds methods for psuedo-NFW profiles to those implemented in lenstronomy"""

    def rho0_c_pseudoNFW(self, c, z):
        """
        computes density normalization as a function of concentration parameter for an NFW-like profile
        with (1+x^2) in the denominator instead of (1+x)^2

        :param c: concentration
        :param z: redshift
        :return: density normalization in h^2/Mpc^3 (physical)
        """
        return 2 * c ** 3 * self.rhoc_z(z) * 200 / (3 * numpy.log(1+c**2))

class LensCosmo(object):

    def __init__(self, z_lens=None, z_source=None, cosmology=None,
                 infall_redshift_model='HYBRID_INFALL', kwargs_infall_model={'log_m_host': 13.0}):
        """

        This class performs calcuations relevant for certain halo mass profiles and lensing-related quantities for a
        given lens/source redshift and cosmology
        :param z_lens: deflector redshift
        :param z_source: source redshift
        :param cosmology: and instance of the Cosmology class (see pyhalo.Cosmology.cosmology.py)
        :param infall_redshift_model: a string or class that determines subhalo infall times
        :param kwargs_infall_model: keyword arguments for infall time model
        """
        if cosmology is None:
            from pyHalo.Cosmology.cosmology import Cosmology
            cosmology = Cosmology()
        self.cosmo = cosmology
        self._arcsec = 2 * numpy.pi / 360 / 3600
        self.h = self.cosmo.h
        # critical density of the universe in M_sun h^2 Mpc^-3
        rhoc = un.Quantity(self.cosmo.astropy.critical_density(0), unit=un.Msun / un.Mpc ** 3).value
        self.rhoc = rhoc / self.cosmo.h ** 2
        if z_lens is not None and z_source is not None:
            # critical density for lensing in units M_sun * Mpc ^ -2
            self.sigma_crit_lensing = self.get_sigma_crit_lensing(z_lens, z_source)
            # critical density for lensing in units M_sun * kpc ^ -2
            self.sigma_crit_lens_kpc = self.sigma_crit_lensing * (0.001) ** 2
            # critical density for lensing in units M_sun * arcsec ^ -2 at lens redshift
            self.sigmacrit = self.sigma_crit_lensing * (0.001) ** 2 * self.cosmo.kpc_proper_per_asec(z_lens) ** 2
            # lensing distances
            self.D_d, self.D_s, self.D_ds = self.cosmo.D_A_z(z_lens), self.cosmo.D_A_z(z_source), self.cosmo.D_A(
                z_lens, z_source)
        self._computed_zacc_pdf = False
        self._nfw_param = NFWParampyHalo(self.cosmo.astropy)
        self.z_lens = z_lens
        self.z_source = z_source
        if infall_redshift_model is not None:
            self.setup_infall_model(infall_redshift_model, kwargs_infall_model)
        else:
            self._infall_pdf_set = False

    def setup_infall_model(self, infall_redshift_model, kwargs_infall_model):

        self._infall_pdf_set = True
        if infall_redshift_model == 'HYBRID_INFALL':
            if 'm_host' in list(kwargs_infall_model.keys()):
                kwargs_infall_model['log_m_host'] = numpy.log10(kwargs_infall_model['m_host'])
                del kwargs_infall_model['m_host']
            if 'log_m_host' not in list(kwargs_infall_model.keys()):
                print('the HYBRID_INFALL model for subhalos requires m_host or log_m_host to be passed as'
                      'keyword arguments through kwargs_infall_model. Using a default value of log_m_host = 13.0')
                kwargs_infall_model['log_m_host'] = 13.0
            self._z_infall_model = InfallDistributionHybrid(self.z_lens, kwargs_infall_model['log_m_host'])
        elif infall_redshift_model == 'DIRECT_INFALL':
            self._z_infall_model = InfallDistributionGalacticus2024(self.z_lens)
        else:
            try:
                self._z_infall_model = infall_redshift_model(self.z_lens, **kwargs_infall_model)
            except:
                if isinstance(infall_redshift_model, str):
                    raise Exception(infall_redshift_model + ' not a valid infall redshift model.')
                else:
                    raise Exception('infall_time_model must be either a class, or a string identifying a '
                                    'particular model. Current options are HYBRID_INFALL and DIRECT_INFALL.')

    def z_accreted_from_zlens(self, m_sub):
        """
        Returns the redshift a subhalo was accreted. Note that in the current implementation this is
        independent of infall mass
        :param m_sub: subhalo mass at infall
        :param z_lens: main deflector redshift
        :return: accretion redshift
        """
        if self._infall_pdf_set:
            return self._z_infall_model(m_sub)
        else:
            raise Exception('must set the infall redshift model before calculating accretion redshift')

    def two_halo_boost(self, m200, z, rmin=0.5, rmax=10):

        """
        Computes the average contribution of the two halo term in a redshift slice adjacent
        the main deflector. Returns a rescaling factor applied to the mass function normalization

        :param m200: host halo mass
        :param z: redshift
        :param rmin: lower limit of the integral, something like the virial radius ~500 kpc
        :param rmax: Upper limit of the integral, this is computed based on redshift spacing during
        the rendering of halos
        :return: scaling factor applied to the normalization of the LOS mass function
        """

        mean_boost = 2 * quad(self.twohaloterm, rmin, rmax, args=(m200, z))[0] / (rmax - rmin)
        # factor of two for symmetry in front/behind host halo

        return 1. + mean_boost

    def twohaloterm(self, r, M, z, mdef='200c'):

        """
        Computes the boost to the background density of the Universe
        from correlated structure around a host of mass M
        :param r:
        :param M:
        :param z:
        :param mdef:
        :return:
        """

        h = self.cosmo.h
        M_h = M * h
        r_h = r * h
        rho_2h = twoHaloTerm(r_h, M_h, z, mdef=mdef) / self.cosmo._colossus_cosmo.rho_m(z)
        return rho_2h

    def nfw_physical2angle(self, m, c, z, pseudo_nfw=False):
        """
        converts the physical mass and concentration parameter of an NFW profile into the lensing quantities
        updates lenstronomy implementation with arbitrary redshift

        :param m: mass enclosed 200 rho_crit in units of M_sun (physical units, meaning no little h)
        :param c: NFW concentration parameter (r200/r_s)
        :param z: redshift at which to evaluate angles from distances
        :param pseudo_nfw: specifies whether one deals with a regualr NFW profile (False) or a psuedo-NFW profile
        with (1+x^2) in the denominator rather than (1+x)^2
        :return: Rs_angle (angle at scale radius) (in units of arcsec), alpha_Rs (observed bending angle at the scale radius
        """
        rho0, Rs, _ = self.nfwParam_physical(m, c, z, pseudo_nfw)
        return self.nfw_physical2angle_fromNFWparams(rho0, Rs, z, pseudo_nfw)

    def nfw_physical2angle_fromNFWparams(self, rho0, Rs, z, pseudo_nfw=False):
        """
        computes the angular lensing properties of an NFW profile from its physical parameters
        :param rho0: central density normalization [M_sun / Mpc^3]
        :param Rs: scale radius [Mpc]
        :param z: redshift at which to evaluate angles from distances
        :param pseudo_nfw: specifies whether one deals with a regular NFW profile (False) or a psuedo-NFW profile
        with (1+x^2) in the denominator rather than (1+x)^2
        :return: scale radius and deflection angle at the scale radius in arcsec
        """
        dd = self.cosmo.D_A_z(z)
        Rs_angle = Rs / dd / self._arcsec  # Rs in arcsec
        if pseudo_nfw:
            r2 = numpy.sqrt(2)
            alpha_Rs = rho0 * (4 * Rs ** 2 * (r2*numpy.log(1+r2) + numpy.log(1. / 2.)))
        else:
            alpha_Rs = rho0 * (4 * Rs ** 2 * (1 + numpy.log(1. / 2.)))
        sigma_crit = self.get_sigma_crit_lensing(z, self.z_source)
        return Rs_angle, alpha_Rs / sigma_crit / dd / self._arcsec

    def rN_M(self, M, z, N):
        """
        computes the radius R_N of a halo of mass M in physical mass M/h, where N is a number multiplying the critical
        density of the Universe at z

        :param M: halo mass in M_sun/h
        :param z: redshift
        :param N: number, e.g. N=200 computes r200
        :return: radius R_N in physical Mpc/h
        """
        rn_mpc_over_h = (3 * M / (4 * numpy.pi * self._nfw_param.rhoc_z(z) * N)) ** (1. / 3.)
        return rn_mpc_over_h / self.cosmo.h

    def nfwParam_physical(self, m, c, z, pseudo_nfw=False):
        """
        returns the NFW parameters in physical units
        updates lenstronomy implementation with arbitrary redshift

        :param m: physical mass in M_sun
        :param c: concentration
        :return: rho0 [Msun/Mpc^3], Rs [Mpc], r200 [Mpc]
        """
        if pseudo_nfw is None:
            raise Exception('psuedo_nfw must be specified in Halo class when accessing nfw parameters!')
        r200 = self._nfw_param.r200_M(m * self.h, z) / self.h  # physical radius r200
        if pseudo_nfw:
            rho0 = self._nfw_param.rho0_c_pseudoNFW(c, z) * self.h ** 2 # physical density in M_sun/Mpc**3
        else:
            rho0 = self._nfw_param.rho0_c(c, z) * self.h**2  # physical density in M_sun/Mpc**3
        Rs = r200/c
        return rho0, Rs, r200

    def NFW_params_physical(self, m, c, z, pseudo_nfw=False):
        """
        returns the NFW parameters in physical units

        :param M: physical mass in M_sun
        :param c: concentration
        :param z: redshift
        :param pseudo_nfw: bool; if False, uses a regular NFW profile, if True, uses an NFW profile
        with (1+x^2) in the denominator
        :return: rho0 [Msun/kpc^3], Rs [kpc], r200 [kpc]
        """
        rho0, Rs, r200 = self.nfwParam_physical(m, c, z, pseudo_nfw)
        return rho0 * 1000 ** -3, Rs * 1000, r200 * 1000

    def sigma_crit_mass(self, z, area):

        """
        :param z: redshift
        :param area: physical area in Mpc^2
        :return: the 'critical density mass' sigma_crit * A in units M_sun
        """

        sigma_crit_mpc = self.get_sigma_crit_lensing(z, self.z_source)
        return area * sigma_crit_mpc

    @property
    def colossus(self):
        return self.cosmo.colossus

    ######################################################
    """ACCESS ROUTINES IN STRUCTURAL PARAMETERS CLASS"""
    ######################################################

    def mthermal_to_halfmode(self, m_thermal):
        """
        Convert thermal relic particle mass to half-mode mass
        :param m_thermaal:
        :return:
        """
        # too lazy for algebra
        def _func(m):
            return abs(self.halfmode_to_thermal(m)-m_thermal)/0.01
        return minimize(_func, x0=10**8, method='Nelder-Mead')['x']

    def halfmode_to_thermal(self, m_half_mode):

        """
        Converts a half mode mass in units of solar masses (no little h) to the mass of
        the corresponding thermal relic particle in keV
        :param m: half mode mass in solar masses
        :return: thermal relic particle mass in keV
        """

        omega_matter = self.cosmo.astropy.Om0
        return 2.32 * (omega_matter / 0.25)**0.4 * (self.cosmo.h/0.7)**0.8 * \
               (m_half_mode / 10 ** 9) ** (-0.3)

    def mhm_to_fsl(self, m_hm):
        """
        Converts half mode mass to free streaming length in Mpc
        See Equations 5-8 in https://arxiv.org/pdf/1112.0330.pdf
        :param m_hm: half-mode mass in units M_sun (no little h)
        :return: free streaming length in units Mpc
        """

        rhoc = self.rhoc * self.cosmo.h ** 2
        l_hm = 2 * (3 * m_hm / (4 * numpy.pi * rhoc)) ** (1. / 3)
        l_fs = l_hm / 13.93
        return l_fs

    ##################################################################################
    """ROUTINES RELATED TO LENSING STUFF"""
    ##################################################################################

    def get_sigma_crit_lensing(self, z1, z2):
        """
        Computes thee critial density for lensing in units of M_sun / Mpc^2
        :param z1: the lens redshit
        :param z2: the source redshift
        :return: the critial density for lensing
        """
        D_ds = self.cosmo.D_A(z1, z2)
        D_d = self.cosmo.D_A_z(z1)
        D_s = self.cosmo.D_A_z(z2)
        d_inv = D_s*D_ds**-1*D_d**-1
        # (Mpc ^2 / sec^2) * (Mpc^-3 M_sun^1 sec ^ 2) * Mpc ^-1 = M_sun / Mpc ^2
        epsilon_crit = (self.cosmo.c**2*(4*numpy.pi*self.cosmo.G)**-1)*d_inv
        return epsilon_crit

    def thetaE_from_sigma(self, z, sigma):
        """
        Calculate the Einstein radius of an SIS profile from velocity dispersion
        :param sigma: velocity dispersion in km/sec
        :return: Einstein radius in arcsec
        """
        arcsec = 206265
        c = 299792.5 # in km/sec
        D_s = self.cosmo.D_A(0.0, self.z_source)
        D_ds = self.cosmo.D_A(z, self.z_source)
        return 4 * numpy.pi * (sigma / c) ** 2 * D_ds / D_s * arcsec

    def point_mass_factor_z(self, z):

        """
        Returns the cosmology-dependent factor to evaluate the Einstein radius of a point mass of mass M:

        :param z: redshift
        :return: The factor that when multiplied by sqrt(mass) gives the Einstein radius of a point mass

        R_ein = sqrt(M) * point_mass_factor_z(z)

        """
        factor = 4 * self.cosmo.G * self.cosmo.c ** -2
        dds = self.cosmo.D_A(z, self.z_source)
        dd = self.cosmo.D_A_z(z)
        ds = self.D_s
        factor *= dds / dd / ds
        return factor ** 0.5 / self.cosmo.arcsec

    def halo_dynamical_time(self, m_host, z, c_host):
        """
        This routine computes the dynamical timescale for a halo of mass M defined as
        t = 0.5427 / sqrt(G*rho)
        where G is the gravitational constant and rho is the average density
        :param m_host: host mass in M_sun
        :param z: host redshift
        :param c_host: host halo concentration
        :return: the dynamical timescale in Gyr
        """

        _, _, rvir = self.NFW_params_physical(m_host, c_host, z)
        volume = (4/3)*numpy.pi*rvir**3
        rho_average = m_host / volume
        g = 4.3e-6
        return 0.5427 / numpy.sqrt(g*rho_average)

    def sidm_halo_effective_age(self, z, z_infall, lambda_t, zform=10.0):
        """
        Calculates a time since z = zform t_1 + t_2 where t_1 is the time from formation to infall, and t_2
        is the time from infall to redshift z times lambda_t
        :param z: halo redshift at the time of lensing
        :param z_infall: infall redshift
        :param lambda_t: rescales the passage of time since the halo becomes a subhalo
        :param zform: formation redshift
        :return: "age" in Gyr
        """
        if z_infall > 10:
            z_infall = 10
        time_formation_to_infall = self.cosmo.halo_age(z_infall, zform=zform)
        time_infall_to_z = self.cosmo.halo_age(z, zform=z_infall)
        return time_formation_to_infall + lambda_t * time_infall_to_z

    def sidm_collapse_timescale(self, rhos, rs, sigma_eff):
        """
        Calculate the SIDM timescale from NFW halo parameters and an effective cross section (Essig et al. 2019)
        :param rhos: NFW halo density scale
        :param rs: NFW halo scale radius
        :param sigma_eff: an "effective" cross section
        :return: timescale in Gyr
        """
        C = 0.75
        G = 4.3e-6 # kpc/M (km/sec)^2
        const1 = 2.0889e-10 # one cm^2 per gram in kpc^2 per M_sun
        const2 = 1.05e-33 # kpc km^2 / s^2 / solar mass in kpc^3 / m_sun / s^2
        denom = const1 * sigma_eff * rhos * rs * numpy.sqrt(4 * numpy.pi * G * rhos * const2)
        tc_seconds = 150 / C / denom
        tc_gyr = tc_seconds * 3.171e-17
        return tc_gyr

    @staticmethod
    def nfw_vmax(rhos, rs):
        """
        Calculate vmax for an NFW profile
        :param rhos: density normalization [solar mass / kpc^3]
        :param rs: scale radius [kpc]
        :return: vmax [km/sec]
        """
        G = 4.3e-6 # kpc / solar mass * (km/sec)^2
        vmax = 1.64 * numpy.sqrt(G * rhos * rs ** 2)
        return vmax



import numpy.testing as npt
import numpy as np
from pyHalo.Halos.HaloModels.NFW_core_trunc import Hybrid, TNFWCHaloEvolving
from pyHalo.Halos.HaloModels.TNFW import TNFWFieldHalo
from pyHalo.truncation_models import ConstantTruncationArcsec
from pyHalo.concentration_models import ConcentrationDiemerJoyce
from pyHalo.Halos.lens_cosmo import LensCosmo
import matplotlib.pyplot as plt
import pytest

class TestTNFWC(object):

    def setup_method(self):

        mass = 10 ** 8
        x = 0.5
        y = 1.0
        z = 0.5
        r3d = 100
        unique_tag = 1.0
        lens_cosmo = LensCosmo(0.5, 2.5)
        kwargs_cdm_profile = {}
        truncation_class = ConstantTruncationArcsec(lens_cosmo, 1000.0)
        concentration_class = ConcentrationDiemerJoyce(lens_cosmo.cosmo, scatter=False)
        self._rescaling_factor = 0.7

        self.tnfw_halo = TNFWFieldHalo(mass, x, y, r3d, z,
                                       False, lens_cosmo, kwargs_cdm_profile,
                                       truncation_class, concentration_class, unique_tag)
        kwargs_profile = {'sidm_timescale': 20.,
                          'lambda_t': 1.0,
                          'mass_conservation': self.tnfw_halo.mass,
                          'rt_kpc': 1000}
        self.tnfwc_halo = TNFWCHaloEvolving(mass, x, y, r3d, z,
                                             False, lens_cosmo, kwargs_profile,
                                             truncation_class, concentration_class, unique_tag)

        self.hybrid = Hybrid(self.tnfw_halo, self.tnfwc_halo, self._rescaling_factor)

    def test_lenstronomy_ID(self):

        ID = self.hybrid.lenstronomy_ID
        npt.assert_string_equal(ID[0],'TNFW')
        npt.assert_string_equal(ID[1], 'TNFWC')

    def test_lenstronomy_args(self):

        kwargs_lenstronomy = self.hybrid.lenstronomy_params[0]
        npt.assert_equal(len(kwargs_lenstronomy),2)

    def test_profile_3d(self):

        rs = self.hybrid.nfw_params[1][1]
        r = np.logspace(-1.5, 1.5, 1000) * rs
        profile_3d_tnfw = self.hybrid.tnfw_halo.density_profile_3d_lenstronomy(r)
        profile_3d_tnfwc = self.hybrid.tnfwc_halo.density_profile_3d_lenstronomy(r)
        plt.loglog(r, profile_3d_tnfw, color='k')
        plt.loglog(r, profile_3d_tnfwc, color='r')
        combined_true = profile_3d_tnfw * (1 - self._rescaling_factor) + profile_3d_tnfwc * self._rescaling_factor
        combined = self.hybrid.density_profile_3d_lenstronomy(r)
        npt.assert_equal(combined, combined_true)

if __name__ == '__main__':
    pytest.main()
